from safeds.ml.nn.typing import ModelImageSize


class DatasetMissesFeaturesError(ValueError):
    """
    Raised when a dataset misses feature columns.

    Parameters
    ----------
    missing_feature_names:
        The names of the missing feature columns.
    """

    def __init__(self, missing_feature_names: list[str]):
        super().__init__(f"Dataset misses the feature columns '{missing_feature_names}'.")


class TargetDataMismatchError(ValueError):
    """
    Raised when the target column of a test dataset mismatches with the target column of the training dataset.

    Currently only used in the Baseline Models.

    Parameters
    ----------
    actual_target_name:
        The actual target column of the dataset.
    missing_target_name:
        The name of the missing target column.
    """

    def __init__(self, actual_target_name: str, missing_target_name: str):
        super().__init__(
            f"The provided target column '{actual_target_name}' does not match the target column of the training set '{missing_target_name}'.",
        )


class DatasetMissesDataError(ValueError):
    """Raised when a dataset contains no rows."""

    def __init__(self) -> None:
        super().__init__("Dataset contains no rows")


class EmptyChoiceError(ValueError):
    """Raised when a choice object is created, but no arguments are provided."""

    def __init__(self) -> None:
        super().__init__("Please provide at least one Value in a Choice Parameter")


class FittingWithChoiceError(Exception):
    """Raised when a model is fitted with a choice object as a parameter."""

    def __init__(self) -> None:
        super().__init__(
            "Error occurred while fitting: Trying to fit with a Choice Parameter. Please use "
            "fit_by_exhaustive_search() instead.",
        )


class FittingWithoutChoiceError(Exception):
    """Raised when a model is fitted by exhaustive search without a choice object as a parameter."""

    def __init__(self) -> None:
        super().__init__(
            "Error occurred while fitting: Trying to fit by exhaustive search without a Choice "
            "Parameter. Please use fit() instead.",
        )


class InvalidFitDataError(Exception):
    """Raised when a Neural Network is fitted on invalid data."""

    def __init__(self, reason: str) -> None:
        super().__init__(f"The given Fit Data is invalid:\n{reason}")


class LearningError(Exception):
    """
    Raised when an error occurred while training a model.

    Parameters
    ----------
    reason:
        The reason for the error.
    """

    def __init__(self, reason: str):
        super().__init__(f"Error occurred while learning: {reason}")


class ModelNotFittedError(RuntimeError):
    """Raised when a model is used before fitting it."""

    def __init__(self) -> None:
        super().__init__("The model has not been fitted yet.")


class InvalidModelStructureError(Exception):
    """Raised when the structure of the model is invalid."""

    def __init__(self, reason: str) -> None:
        super().__init__(f"The model structure is invalid: {reason}")


class PredictionError(Exception):
    """
    Raised when an error occurred while prediction a target vector using a model.

    Parameters
    ----------
    reason:
        The reason for the error.
    """

    def __init__(self, reason: str):
        super().__init__(f"Error occurred while predicting: {reason}")


class FeatureDataMismatchError(Exception):
    """Raised when the columns of the table passed to the predict or fit method do not match with the specified features of the model."""

    def __init__(self) -> None:
        super().__init__(
            "The features in the given table do not match with the specified feature columns names of the model.",
        )


class InputSizeError(Exception):
    """Raised when the amount of features being passed to a model does not match with its input size."""

    def __init__(self, data_size: int | ModelImageSize, input_layer_size: int | ModelImageSize | None) -> None:
        # TODO: remove input_layer_size type None again
        super().__init__(
            f"The data size being passed to the network({data_size}) does not match with its input size({input_layer_size}). Consider changing the data size of the model or reformatting the data.",
        )


class PlainTableError(TypeError):
    """Raised when a plain table is used instead of a TabularDataset."""

    def __init__(self) -> None:
        super().__init__(
            (
                "This method needs a tabular dataset. "
                "It knows which columns are features and which are the target to predict.\n"
                "Use `Table.to_tabular_dataset()` to create a tabular dataset."
            ),
        )



from __future__ import annotations

import io
import itertools
from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds.data.image.containers import Image
from safeds.data.tabular.containers import Column
from safeds.exceptions import (
    DatasetMissesDataError,
    MissingValuesColumnError,
    ModelNotFittedError,
    NonNumericColumnError,
)

if TYPE_CHECKING:
    from statsmodels.tsa.arima.model import ARIMA

    from safeds.data.labeled.containers import TimeSeriesDataset
    from safeds.data.tabular.containers import Table


class ArimaModelRegressor:
    """Auto Regressive Integrated Moving Average Model."""

    def __hash__(self) -> int:
        """
        Return a deterministic hash value for a regressor.

        Returns
        -------
        hash:
            The hash value.
        """
        return _structural_hash(self.__class__.__qualname__, self.is_fitted, self._order)

    def __init__(self) -> None:
        # Internal state
        self._arima: ARIMA | None = None
        self._order: tuple[int, int, int] | None = None
        self._fitted = False

    def fit(self, time_series: TimeSeriesDataset) -> ArimaModelRegressor:
        """
        Create a copy of this ARIMA Model and fit it with the given training data.

        This ARIMA Model is not modified.

        Parameters
        ----------
        time_series:
            The time series containing the target column, which will be used.

        Returns
        -------
        fitted_arima:
            The fitted ARIMA Model.

        Raises
        ------
        LearningError
            If the training data contains invalid values or if the training failed.
        NonTimeSeriesError
            If the table is not a TimeSeries object.
        NonNumericColumnError
            If the training data contains non-numerical values.
        MissingValuesColumnError
            If the training data contains missing values.
        DatasetMissesDataError
            If the training data contains no rows.
        """
        from statsmodels.tsa.arima.model import ARIMA

        table = time_series.to_table()
        if table.row_count == 0:
            raise DatasetMissesDataError
        if not time_series.target.type.is_numeric:
            raise NonNumericColumnError(time_series.target.name)
        if time_series.target.missing_value_count() > 0:
            raise MissingValuesColumnError(
                time_series.target.name,
                "You can use the Imputer to replace the missing values based on different strategies.\nIf you want to"
                "remove the missing values entirely you can use the method "
                "`TimeSeries.remove_rows_with_missing_values`.",
            )
        fitted_arima = ArimaModelRegressor()
        p = d = q = range(2)
        pdq = list(itertools.product(p, d, q))
        best_aic = float("inf")
        best_model = None
        # best param will get overwritten
        best_param = (0, 0, 0)
        for param in pdq:
            # Create and fit an ARIMA model with the current parameters
            mod = ARIMA(time_series.target._series.to_numpy(), order=param)

            # I wasnt able to invoke an learning Error
            # Add try catch when an learning error is found
            result = mod.fit()

            # Compare the current model's AIC with the best AIC so far
            if result.aic < best_aic:
                best_aic = result.aic
                best_param = param
                best_model = result

        fitted_arima._order = best_param
        fitted_arima._arima = best_model
        fitted_arima._fitted = True
        return fitted_arima

    def predict(self, time_series: TimeSeriesDataset) -> Table:
        """
        Predict a target vector using a time series target column. The model has to be trained first.

        Parameters
        ----------
        time_series:
            The test dataset of the time series.

        Returns
        -------
        time_series:
            A timeseries containing the predicted target vector and a time dummy as time column.

        Raises
        ------
        ModelNotFittedError
            If the model has not been fitted yet.
        IndexError
            If the forecast horizon is not greater than zero.
        PredictionError
            If predicting with the given dataset failed.
        """
        # make a table without
        forecast_horizon = len(time_series.target._series.to_numpy())
        result_table = time_series.to_table()
        result_table = result_table.remove_columns([time_series.target.name], ignore_unknown_names=True)
        # Validation
        if not self.is_fitted or self._arima is None:
            raise ModelNotFittedError

        # forecast
        # couldn't invoke prediction error, will be added when found
        forecast_results = self._arima.forecast(steps=forecast_horizon)
        target_column: Column = Column(name=time_series.target.name, data=forecast_results)

        # create new TimeSeries
        return result_table.add_columns(target_column)

    def plot_predictions(self, test_series: TimeSeriesDataset) -> Image:
        """
        Plot the predictions of the trained model to the given target of the time series. So you can see the predictions and the actual values in one plot.

        Parameters
        ----------
        test_series:
            The time series containing the target vector.

        Returns
        -------
        image:
            Plots predictions of the given time series to the given target Column

        Raises
        ------
        ModelNotFittedError
            If the model has not been fitted yet.
        PredictionError
            If predicting with the given dataset failed.

        """
        import matplotlib.pyplot as plt

        if not self.is_fitted or self._arima is None:
            raise ModelNotFittedError
        test_data = test_series.target._series.to_numpy()
        n_steps = len(test_data)
        forecast_results = self._arima.forecast(steps=n_steps)

        fig = plt.figure()
        plt.plot(
            forecast_results,
        )
        plt.legend(["forecasted"])
        plt.plot(test_data)
        plt.tight_layout()

        buffer = io.BytesIO()
        fig.savefig(buffer, format="png")
        plt.close()  # Prevents the figure from being displayed directly
        buffer.seek(0)
        return Image.from_bytes(buffer.read())

    @property
    def is_fitted(self) -> bool:
        """Whether the regressor is fitted."""
        return self._fitted



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from safeds.data.image.containers import ImageList
from safeds.data.labeled.containers import ImageDataset, TabularDataset, TimeSeriesDataset
from safeds.data.tabular.containers import Table

if TYPE_CHECKING:
    from torch import Tensor
    from torch.utils.data import DataLoader

    from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
    from safeds.ml.nn.typing import ModelImageSize

FT = TypeVar("FT", TabularDataset, TimeSeriesDataset, ImageDataset)
PT = TypeVar("PT", Table, TimeSeriesDataset, ImageList)


class InputConversion(Generic[FT, PT], ABC):
    """The input conversion for a neural network, defines the input parameters for the neural network."""

    @property
    @abstractmethod
    def _data_size(self) -> int | ModelImageSize: ...

    @abstractmethod
    def _data_conversion_fit(
        self,
        input_data: FT,
        batch_size: int,
        num_of_classes: int = 1,
    ) -> DataLoader | ImageDataset: ...

    @abstractmethod
    def _data_conversion_predict(self, input_data: PT, batch_size: int) -> DataLoader | _SingleSizeImageList: ...

    @abstractmethod
    def _data_conversion_output(self, input_data: PT, output_data: Tensor) -> FT: ...

    @abstractmethod
    def _is_fit_data_valid(self, input_data: FT) -> bool: ...

    @abstractmethod
    def _is_predict_data_valid(self, input_data: PT) -> bool: ...



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator, Mapping
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from safeds.data.tabular.typing import DataType, Schema

    from ._cell import Cell


class Row(ABC, Mapping[str, Any]):
    """
    A one-dimensional collection of named, heterogeneous values.

    This class cannot be instantiated directly. It is only used for arguments of callbacks.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __contains__(self, name: Any) -> bool:
        return self.has_column(name)

    @abstractmethod
    def __eq__(self, other: object) -> bool: ...

    def __getitem__(self, name: str) -> Cell:
        return self.get_value(name)

    @abstractmethod
    def __hash__(self) -> int: ...

    def __iter__(self) -> Iterator[Any]:
        return iter(self.column_names)

    def __len__(self) -> int:
        return self.column_count

    @abstractmethod
    def __sizeof__(self) -> int: ...

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def column_names(self) -> list[str]:
        """The names of the columns in the row."""

    @property
    @abstractmethod
    def column_count(self) -> int:
        """The number of columns in the row."""

    @property
    @abstractmethod
    def schema(self) -> Schema:
        """The schema of the row."""

    # ------------------------------------------------------------------------------------------------------------------
    # Column operations
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def get_value(self, name: str) -> Cell:
        """
        Get the value of the specified column. This is equivalent to using the `[]` operator (indexed access).

        Parameters
        ----------
        name:
            The name of the column.

        Returns
        -------
        value:
            The value of the column.

        Raises
        ------
        ColumnNotFoundError
            If the column name does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"col1": [1, 2], "col2": [3, 4]})
        >>> table.remove_rows(lambda row: row.get_value("col1") == 1)
        +------+------+
        | col1 | col2 |
        |  --- |  --- |
        |  i64 |  i64 |
        +=============+
        |    2 |    4 |
        +------+------+


        >>> table.remove_rows(lambda row: row["col1"] == 1)
        +------+------+
        | col1 | col2 |
        |  --- |  --- |
        |  i64 |  i64 |
        +=============+
        |    2 |    4 |
        +------+------+
        """

    @abstractmethod
    def get_column_type(self, name: str) -> DataType:
        """
        Get the type of the specified column.

        Parameters
        ----------
        name:
            The name of the column.

        Returns
        -------
        type:
            The type of the column.

        Raises
        ------
        ColumnNotFoundError
            If the column name does not exist.
        """

    @abstractmethod
    def has_column(self, name: str) -> bool:
        """
        Check if the row has a column with the specified name.

        Parameters
        ----------
        name:
            The name of the column.

        Returns
        -------
        has_column:
            Whether the row has a column with the specified name.
        """



from __future__ import annotations

from typing import TYPE_CHECKING

from ._data_type import DataType

if TYPE_CHECKING:
    import polars as pl


class _PolarsDataType(DataType):
    """
    The type of a column or cell in a table.

    This implementation is based on Polars' data types.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, dtype: pl.DataType):
        self._dtype: pl.DataType = dtype

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _PolarsDataType):
            return NotImplemented
        if self is other:
            return True
        return self._dtype.is_(other._dtype)

    def __hash__(self) -> int:
        return self._dtype.__hash__()

    def __repr__(self) -> str:
        return self._dtype.__repr__()

    def __sizeof__(self) -> int:
        return self._dtype.__sizeof__()

    def __str__(self) -> str:
        return self._dtype.__str__()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def is_numeric(self) -> bool:
        return self._dtype.is_numeric()

    @property
    def is_temporal(self) -> bool:
        return self._dtype.is_temporal()



"""Converters between our data contains and tensors."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._input_converter import InputConversion
    from ._input_converter_image_to_column import InputConversionImageToColumn
    from ._input_converter_image_to_image import InputConversionImageToImage
    from ._input_converter_image_to_table import InputConversionImageToTable
    from ._input_converter_table import InputConversionTable
    from ._input_converter_time_series import InputConversionTimeSeries

apipkg.initpkg(
    __name__,
    {
        "InputConversion": "._input_converter:InputConversion",
        "InputConversionImageToColumn": "._input_converter_image_to_column:InputConversionImageToColumn",
        "InputConversionImageToImage": "._input_converter_image_to_image:InputConversionImageToImage",
        "InputConversionImageToTable": "._input_converter_image_to_table:InputConversionImageToTable",
        "InputConversionTable": "._input_converter_table:InputConversionTable",
        "InputConversionTimeSeries": "._input_converter_time_series:InputConversionTimeSeries",
    },
)

__all__ = [
    "InputConversion",
    "InputConversionImageToColumn",
    "InputConversionImageToImage",
    "InputConversionImageToTable",
    "InputConversionTable",
    "InputConversionTimeSeries",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._validation import _check_columns_exist

from ._lazy_cell import _LazyCell
from ._row import Row

if TYPE_CHECKING:
    from safeds.data.tabular.typing import DataType, Schema

    from ._table import Table


class _LazyVectorizedRow(Row):
    """
    A one-dimensional collection of named, heterogeneous values.

    This implementation treats an entire table as a row, where each column is a "cell" in the row. This greatly speeds
    up operations on the row.

    Moreover, accessing a column only builds an expression that will be evaluated when needed. This is useful when later
    operations remove more rows or columns, so we don't do unnecessary work upfront.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, table: Table):
        self._table: Table = table

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _LazyVectorizedRow):
            return NotImplemented
        if self is other:
            return True
        return self._table == other._table

    def __hash__(self) -> int:
        return self._table.__hash__()

    def __sizeof__(self) -> int:
        return self._table.__sizeof__()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def column_names(self) -> list[str]:
        return self._table.column_names

    @property
    def column_count(self) -> int:
        return self._table.column_count

    @property
    def schema(self) -> Schema:
        return self._table.schema

    # ------------------------------------------------------------------------------------------------------------------
    # Column operations
    # ------------------------------------------------------------------------------------------------------------------

    def get_value(self, name: str) -> _LazyCell:
        import polars as pl

        _check_columns_exist(self._table, name)

        return _LazyCell(pl.col(name))

    def get_column_type(self, name: str) -> DataType:
        return self._table.get_column_type(name)

    def has_column(self, name: str) -> bool:
        return self._table.has_column(name)



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from torch.types import Device


_default_device: Device | None = None


def _get_device() -> Device:
    import torch

    return torch.get_default_device()


def _init_default_device() -> None:
    import torch

    global _default_device  # noqa: PLW0603

    if _default_device is None:
        _default_device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")

    torch.set_default_device(_default_device)


def _set_default_device(device: Device) -> None:
    # This changes all future tensors, but not any tensor that already exists
    global _default_device  # noqa: PLW0603

    _default_device = device
    _init_default_device()



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound

from ._lazy_cell import _LazyCell
from ._string_cell import StringCell

if TYPE_CHECKING:
    import datetime

    import polars as pl

    from ._cell import Cell


class _LazyStringCell(StringCell):
    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, expression: pl.Expr) -> None:
        self._expression: pl.Expr = expression

    def __hash__(self) -> int:
        return _structural_hash(self._expression.meta.serialize())

    def __sizeof__(self) -> int:
        return self._expression.__sizeof__()

    # ------------------------------------------------------------------------------------------------------------------
    # String operations
    # ------------------------------------------------------------------------------------------------------------------

    def contains(self, substring: str) -> Cell[bool]:
        return _LazyCell(self._expression.str.contains(substring, literal=True))

    def length(self, optimize_for_ascii: bool = False) -> Cell[int]:
        if optimize_for_ascii:
            return _LazyCell(self._expression.str.len_bytes())
        else:
            return _LazyCell(self._expression.str.len_chars())

    def ends_with(self, suffix: str) -> Cell[bool]:
        return _LazyCell(self._expression.str.ends_with(suffix))

    def index_of(self, substring: str) -> Cell[int | None]:
        return _LazyCell(self._expression.str.find(substring, literal=True))

    def replace(self, old: str, new: str) -> Cell[str]:
        return _LazyCell(self._expression.str.replace_all(old, new, literal=True))

    def starts_with(self, prefix: str) -> Cell[bool]:
        return _LazyCell(self._expression.str.starts_with(prefix))

    def substring(self, start: int = 0, length: int | None = None) -> Cell[str]:
        _check_bounds("length", length, lower_bound=_ClosedBound(0))

        return _LazyCell(self._expression.str.slice(start, length))

    def to_date(self) -> Cell[datetime.date | None]:
        return _LazyCell(self._expression.str.to_date(format="%F", strict=False))

    def to_datetime(self) -> Cell[datetime.datetime | None]:
        return _LazyCell(self._expression.str.to_datetime(format="%+", strict=False))

    def to_int(self, *, base: int = 10) -> Cell[int | None]:
        return _LazyCell(self._expression.str.to_integer(base=base, strict=False))

    def to_float(self) -> Cell[float | None]:
        import polars as pl

        return _LazyCell(self._expression.cast(pl.Float64, strict=False))

    def to_lowercase(self) -> Cell[str]:
        return _LazyCell(self._expression.str.to_lowercase())

    def to_uppercase(self) -> Cell[str]:
        return _LazyCell(self._expression.str.to_uppercase())

    def trim(self) -> Cell[str]:
        return _LazyCell(self._expression.str.strip_chars())

    def trim_end(self) -> Cell[str]:
        return _LazyCell(self._expression.str.strip_chars_end())

    def trim_start(self) -> Cell[str]:
        return _LazyCell(self._expression.str.strip_chars_start())

    # ------------------------------------------------------------------------------------------------------------------
    # Internal
    # ------------------------------------------------------------------------------------------------------------------

    def _equals(self, other: object) -> bool:
        if not isinstance(other, _LazyStringCell):
            return NotImplemented
        if self is other:
            return True
        return self._expression.meta.eq(other._expression.meta)



from safeds.ml.nn.typing import ConstantImageSize


class ImageSize(ConstantImageSize):
    """
    A container for image size data.

    Parameters
    ----------
    width:
        the width of the image
    height:
        the height of the image
    channel:
        the channel of the image

    Raises
    ------
    OutOfBoundsError:
        if width or height are below 1
    ValueError
        if an invalid channel is given
    """

    def __str__(self) -> str:
        return f"{self._width}x{self._height}x{self._channel} (WxHxC)"



"""Neural networks for various tasks."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._model import NeuralNetworkClassifier, NeuralNetworkRegressor

apipkg.initpkg(
    __name__,
    {
        "NeuralNetworkClassifier": "._model:NeuralNetworkClassifier",
        "NeuralNetworkRegressor": "._model:NeuralNetworkRegressor",
    },
)

__all__ = [
    "NeuralNetworkClassifier",
    "NeuralNetworkRegressor",
]



from __future__ import annotations

import copy
import sys
import warnings
from typing import TYPE_CHECKING, TypeVar

from safeds._config import _get_device, _init_default_device
from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound
from safeds.data.image.containers import ImageList
from safeds.data.image.containers._empty_image_list import _EmptyImageList
from safeds.data.image.containers._multi_size_image_list import _MultiSizeImageList
from safeds.data.image.containers._single_size_image_list import _SingleSizeImageList
from safeds.data.image.typing import ImageSize
from safeds.data.tabular.containers import Column, Table
from safeds.data.tabular.transformation import OneHotEncoder
from safeds.exceptions import (
    IndexOutOfBoundsError,
    NonNumericColumnError,
    OutputLengthMismatchError,
    TransformerNotFittedError,
)

from ._dataset import Dataset

if TYPE_CHECKING:
    from torch import Tensor

Out_co = TypeVar("Out_co", Column, ImageList, Table, covariant=True)


class ImageDataset(Dataset[ImageList, Out_co]):
    """
    A Dataset for ImageLists as input and ImageLists, Tables or Columns as output.

    Parameters
    ----------
    input_data:
        the input ImageList
    output_data:
        the output data
    batch_size:
        the batch size used for training
    shuffle:
        whether the data should be shuffled after each epoch of training
    """

    def __init__(self, input_data: ImageList, output_data: Out_co, batch_size: int = 1, shuffle: bool = False) -> None:
        import torch

        _init_default_device()

        self._shuffle_tensor_indices: torch.Tensor = torch.tensor(list(range(len(input_data))), dtype=torch.int64)
        self._shuffle_after_epoch: bool = shuffle
        self._batch_size: int = batch_size
        self._next_batch_index: int = 0

        if isinstance(input_data, _MultiSizeImageList):
            raise ValueError("The given input ImageList contains images of different sizes.")  # noqa: TRY004
        elif isinstance(input_data, _EmptyImageList):
            raise ValueError("The given input ImageList contains no images.")  # noqa: TRY004
        else:
            self._input_size: ImageSize = ImageSize(input_data.widths[0], input_data.heights[0], input_data.channel)
            self._input: _SingleSizeImageList = input_data._as_single_size_image_list()
        if ((isinstance(output_data, Column | Table)) and len(input_data) != output_data.row_count) or (
            isinstance(output_data, ImageList) and len(input_data) != len(output_data)
        ):
            if isinstance(output_data, Table):
                output_len = output_data.row_count
            else:
                output_len = len(output_data)
            raise OutputLengthMismatchError(f"{len(input_data)} != {output_len}")
        if isinstance(output_data, Table):
            non_numerical_columns = []
            wrong_interval_columns = []
            for column_name in output_data.column_names:
                if not output_data.get_column_type(column_name).is_numeric:
                    non_numerical_columns.append(column_name)
                elif (output_data.get_column(column_name).min() or 0) < 0 or (
                    output_data.get_column(column_name).max() or 0
                ) > 1:
                    wrong_interval_columns.append(column_name)
            if len(non_numerical_columns) > 0:
                raise NonNumericColumnError(f"Columns {non_numerical_columns} are not numerical.")
            if len(wrong_interval_columns) > 0:
                raise ValueError(f"Columns {wrong_interval_columns} have values outside of the interval [0, 1].")
            _output: _TableAsTensor | _ColumnAsTensor | _SingleSizeImageList = _TableAsTensor(output_data)
            _output_size: int | ImageSize = output_data.column_count
        elif isinstance(output_data, Column):
            _column_as_tensor = _ColumnAsTensor(output_data)
            _output_size = len(_column_as_tensor._one_hot_encoder._get_names_of_added_columns())
            _output = _column_as_tensor
        elif isinstance(output_data, _SingleSizeImageList):
            _output = output_data._clone()._as_single_size_image_list()
            _output_size = ImageSize(output_data.widths[0], output_data.heights[0], output_data.channel)
        else:
            raise ValueError("The given output ImageList contains images of different sizes.")  # noqa: TRY004
        self._output = _output  # type: ignore[var-annotated]  # TODO: check what the type should be
        self._output_size = _output_size  # type: ignore[var-annotated]  # TODO: check what the type should be

    def __iter__(self) -> ImageDataset:
        if self._shuffle_after_epoch:
            im_ds = self.shuffle()
        else:
            im_ds = copy.copy(self)
        im_ds._next_batch_index = 0
        return im_ds

    def __next__(self) -> tuple[Tensor, Tensor]:
        if self._next_batch_index * self._batch_size >= len(self._shuffle_tensor_indices):
            raise StopIteration
        self._next_batch_index += 1
        return self._get_batch(self._next_batch_index - 1)

    def __len__(self) -> int:
        return len(self._shuffle_tensor_indices)

    def __eq__(self, other: object) -> bool:
        """
        Compare two image datasets.

        Parameters
        ----------
        other:
            The image dataset to compare to.

        Returns
        -------
        equals:
            Whether the two image datasets are the same.
        """
        if not isinstance(other, ImageDataset):
            return NotImplemented
        return (self is other) or (
            self._shuffle_after_epoch == other._shuffle_after_epoch
            and self._batch_size == other._batch_size
            and isinstance(other._output, type(self._output))
            and (self._input == other._input)
            and (self._output == other._output)
            and (self._shuffle_tensor_indices.tolist() == other._shuffle_tensor_indices.tolist())
        )

    def __hash__(self) -> int:
        """
        Return a deterministic hash value for this image dataset.

        Returns
        -------
        hash:
            the hash value
        """
        return _structural_hash(
            self._input,
            self._output,
            self._shuffle_after_epoch,
            self._batch_size,
            self._shuffle_tensor_indices.tolist(),
        )

    def __sizeof__(self) -> int:
        """
        Return the complete size of this object.

        Returns
        -------
        size:
            Size of this object in bytes.
        """
        return (
            sys.getsizeof(self._shuffle_tensor_indices)
            + self._shuffle_tensor_indices.element_size() * self._shuffle_tensor_indices.nelement()
            + sys.getsizeof(self._input)
            + sys.getsizeof(self._output)
            + sys.getsizeof(self._input_size)
            + sys.getsizeof(self._output_size)
            + sys.getsizeof(self._shuffle_after_epoch)
            + sys.getsizeof(self._batch_size)
            + sys.getsizeof(self._next_batch_index)
        )

    @property
    def input_size(self) -> ImageSize:
        """
        Get the input `ImageSize` of this dataset.

        Returns
        -------
        input_size:
            the input `ImageSize`
        """
        return self._input_size

    @property
    def output_size(self) -> ImageSize | int:
        """
        Get the output size of this dataset.

        Returns
        -------
        output_size:
            the output size
        """
        return self._output_size

    def get_input(self) -> ImageList:
        """
        Get the input data of this dataset.

        Returns
        -------
        input:
            the input data of this dataset
        """
        return self._sort_image_list_with_shuffle_tensor_indices_reduce_if_necessary(self._input)

    def get_output(self) -> Out_co:
        """
        Get the output data of this dataset.

        Returns
        -------
        output:
            the output data of this dataset
        """
        output = self._output
        if isinstance(output, _TableAsTensor):
            return output._to_table(self._shuffle_tensor_indices)  # type: ignore[return-value]
        elif isinstance(output, _ColumnAsTensor):
            return output._to_column(self._shuffle_tensor_indices)  # type: ignore[return-value]
        else:
            return self._sort_image_list_with_shuffle_tensor_indices_reduce_if_necessary(self._output)  # type: ignore[return-value]

    def _sort_image_list_with_shuffle_tensor_indices_reduce_if_necessary(
        self,
        image_list: _SingleSizeImageList,
    ) -> _SingleSizeImageList:
        shuffled_image_list = _SingleSizeImageList()
        tensor_pos = [
            image_list._indices_to_tensor_positions[shuffled_index]
            for shuffled_index in sorted(self._shuffle_tensor_indices.tolist())
        ]
        temp_pos = {
            shuffled_index: new_index for new_index, shuffled_index in enumerate(self._shuffle_tensor_indices.tolist())
        }
        shuffled_image_list._tensor = image_list._tensor[tensor_pos]
        shuffled_image_list._tensor_positions_to_indices = [
            new_index for _, new_index in sorted(temp_pos.items(), key=lambda item: item[0])
        ]
        shuffled_image_list._indices_to_tensor_positions = shuffled_image_list._calc_new_indices_to_tensor_positions()
        return shuffled_image_list

    def _get_batch(self, batch_number: int, batch_size: int | None = None) -> tuple[Tensor, Tensor]:
        import torch

        _init_default_device()

        if batch_size is None:
            batch_size = self._batch_size

        _check_bounds("batch_size", batch_size, lower_bound=_ClosedBound(1))

        if batch_number < 0 or batch_size * batch_number >= len(self._shuffle_tensor_indices):
            raise IndexOutOfBoundsError(batch_size * batch_number)
        max_index = (
            batch_size * (batch_number + 1)
            if batch_size * (batch_number + 1) < len(self._shuffle_tensor_indices)
            else len(self._shuffle_tensor_indices)
        )
        input_tensor = (
            self._input._tensor[
                [
                    self._input._indices_to_tensor_positions[index]
                    for index in self._shuffle_tensor_indices[batch_size * batch_number : max_index].tolist()
                ]
            ].to(torch.float32)
            / 255
        )
        output_tensor: Tensor
        if isinstance(self._output, _SingleSizeImageList):
            output_tensor = (
                self._output._tensor[
                    [
                        self._input._indices_to_tensor_positions[index]
                        for index in self._shuffle_tensor_indices[batch_size * batch_number : max_index].tolist()
                    ]
                ].to(torch.float32)
                / 255
            )
        else:  # _output is instance of _TableAsTensor or _ColumnAsTensor
            output_tensor = self._output._tensor[self._shuffle_tensor_indices[batch_size * batch_number : max_index]]
        return input_tensor, output_tensor

    def shuffle(self) -> ImageDataset[Out_co]:
        """
        Return a new `ImageDataset` with shuffled data.

        The original dataset is not modified.

        Returns
        -------
        image_dataset:
            the shuffled `ImageDataset`
        """
        import torch

        _init_default_device()

        im_dataset: ImageDataset[Out_co] = copy.copy(self)
        im_dataset._shuffle_tensor_indices = self._shuffle_tensor_indices[
            torch.randperm(len(self._shuffle_tensor_indices))
        ]
        im_dataset._next_batch_index = 0
        return im_dataset

    def split(
        self,
        percentage_in_first: float,
        *,
        shuffle: bool = True,
    ) -> tuple[ImageDataset[Out_co], ImageDataset[Out_co]]:
        """
        Create two image datasets by splitting the data of the current dataset.

        The first dataset contains a percentage of the data specified by `percentage_in_first`, and the second dataset
        contains the remaining data.

        The original dataset is not modified.
        By default, the data is shuffled before splitting. You can disable this by setting `shuffle` to False.

        Parameters
        ----------
        percentage_in_first:
            The percentage of data to include in the first dataset. Must be between 0 and 1.
        shuffle:
            Whether to shuffle the data before splitting.

        Returns
        -------
        first_dataset:
            The first dataset.
        second_dataset:
            The second dataset.

        Raises
        ------
        OutOfBoundsError
            If `percentage_in_first` is not between 0 and 1.
        """
        import torch

        _check_bounds(
            "percentage_in_first",
            percentage_in_first,
            lower_bound=_ClosedBound(0),
            upper_bound=_ClosedBound(1),
        )

        first_dataset: ImageDataset[Out_co] = copy.copy(self)
        second_dataset: ImageDataset[Out_co] = copy.copy(self)

        if shuffle:
            shuffled_indices = torch.randperm(len(self._shuffle_tensor_indices))
        else:
            shuffled_indices = torch.arange(len(self._shuffle_tensor_indices))

        first_dataset._shuffle_tensor_indices, second_dataset._shuffle_tensor_indices = shuffled_indices.split(
            [
                round(percentage_in_first * len(self)),
                len(self) - round(percentage_in_first * len(self)),
            ],
        )
        return first_dataset, second_dataset


class _TableAsTensor:
    def __init__(self, table: Table) -> None:
        import polars as pl
        import torch

        _init_default_device()

        self._column_names = table.column_names
        if table.row_count == 0:
            self._tensor = torch.empty((0, table.column_count), dtype=torch.float32).to(_get_device())
        else:
            self._tensor = table._data_frame.to_torch(dtype=pl.Float32).to(_get_device())

        if not torch.all(self._tensor.sum(dim=1) == torch.ones(self._tensor.size(dim=0))):
            raise ValueError(
                "The given table is not correctly one hot encoded as it contains rows that have a sum not equal to 1.",
            )

    def __eq__(self, other: object) -> bool:
        import torch

        _init_default_device()

        if not isinstance(other, _TableAsTensor):
            return NotImplemented
        return (self is other) or (
            self._column_names == other._column_names and torch.all(torch.eq(self._tensor, other._tensor)).item()
        )

    def __hash__(self) -> int:
        return _structural_hash(self._tensor.size(), self._column_names)

    def __sizeof__(self) -> int:
        return (
            sys.getsizeof(self._tensor)
            + self._tensor.element_size() * self._tensor.nelement()
            + sys.getsizeof(self._column_names)
        )

    @staticmethod
    def _from_tensor(tensor: Tensor, column_names: list[str]) -> _TableAsTensor:
        if tensor.dim() != 2:
            raise ValueError(f"Tensor has an invalid amount of dimensions. Needed 2 dimensions but got {tensor.dim()}.")
        if tensor.size(dim=1) != len(column_names):
            raise ValueError(
                f"Tensor and column_names have different amounts of classes ({tensor.size(dim=1)}!={len(column_names)}).",
            )
        table_as_tensor = _TableAsTensor.__new__(_TableAsTensor)
        table_as_tensor._tensor = tensor
        table_as_tensor._column_names = column_names
        return table_as_tensor

    def _to_table(self, shuffled_indices: Tensor) -> Table:
        return Table(dict(zip(self._column_names, self._tensor[shuffled_indices].T.tolist(), strict=False)))


class _ColumnAsTensor:
    def __init__(self, column: Column) -> None:
        import polars as pl
        import torch

        _init_default_device()

        self._column_name = column.name
        column_as_table = Table.from_columns([column])
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=rf"The columns \['{self._column_name}'\] contain numerical data. The OneHotEncoder is designed to encode non-numerical values into numerical values",
                category=UserWarning,
            )
            # TODO: should not one-hot-encode the target. label encoding without order is sufficient. should also not
            #  be done automatically?
            self._one_hot_encoder = OneHotEncoder(column_names=self._column_name).fit(column_as_table)
        self._tensor = torch.Tensor(
            self._one_hot_encoder.transform(column_as_table)._data_frame.to_torch(dtype=pl.Float32),
        ).to(_get_device())

    def __eq__(self, other: object) -> bool:
        import torch

        _init_default_device()

        if not isinstance(other, _ColumnAsTensor):
            return NotImplemented
        return (self is other) or (
            self._column_name == other._column_name
            and self._one_hot_encoder == other._one_hot_encoder
            and torch.all(torch.eq(self._tensor, other._tensor)).item()
        )

    def __hash__(self) -> int:
        return _structural_hash(self._tensor.size(), self._column_name, self._one_hot_encoder)

    def __sizeof__(self) -> int:
        return (
            sys.getsizeof(self._tensor)
            + self._tensor.element_size() * self._tensor.nelement()
            + sys.getsizeof(self._column_name)
            + sys.getsizeof(self._one_hot_encoder)
        )

    @staticmethod
    def _from_tensor(tensor: Tensor, column_name: str, one_hot_encoder: OneHotEncoder) -> _ColumnAsTensor:
        if tensor.dim() != 2:
            raise ValueError(f"Tensor has an invalid amount of dimensions. Needed 2 dimensions but got {tensor.dim()}.")
        if not one_hot_encoder.is_fitted:
            raise TransformerNotFittedError
        if tensor.size(dim=1) != len(one_hot_encoder._get_names_of_added_columns()):
            raise ValueError(
                f"Tensor and one_hot_encoder have different amounts of classes ({tensor.size(dim=1)}!={len(one_hot_encoder._get_names_of_added_columns())}).",
            )
        table_as_tensor = _ColumnAsTensor.__new__(_ColumnAsTensor)
        table_as_tensor._tensor = tensor
        table_as_tensor._column_name = column_name
        table_as_tensor._one_hot_encoder = one_hot_encoder
        return table_as_tensor

    def _to_column(self, shuffled_indices: Tensor) -> Column:
        table = Table(
            dict(
                zip(
                    self._one_hot_encoder._get_names_of_added_columns(),
                    self._tensor[shuffled_indices].T.tolist(),
                    strict=False,
                ),
            ),
        )
        return self._one_hot_encoder.inverse_transform(table).get_column(self._column_name)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds._validation import _check_bounds, _ClosedBound, _OpenBound

if TYPE_CHECKING:
    from safeds.ml.classical import SupervisedModel


class _AdaBoostBase(ABC):
    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def __init__(
        self,
        max_learner_count: int,
        learning_rate: float,
    ) -> None:
        # Validation
        _check_bounds("max_learner_count", max_learner_count, lower_bound=_ClosedBound(1))
        _check_bounds("learning_rate", learning_rate, lower_bound=_OpenBound(0))

        # Hyperparameters
        self._max_learner_count: int = max_learner_count
        self._learning_rate: float = learning_rate

    def __hash__(self) -> int:
        return _structural_hash(
            self._max_learner_count,
            self._learning_rate,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def max_learner_count(self) -> int:
        """The maximum number of learners in the ensemble."""
        return self._max_learner_count

    @property
    def learning_rate(self) -> float:
        """The learning rate."""
        return self._learning_rate

    @property
    @abstractmethod
    def learner(self) -> SupervisedModel | None:
        """The base learner used for training the ensemble."""



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds.data.image.containers import Image
from safeds.exceptions._ml import ModelNotFittedError
from safeds.ml.classical._bases import _DecisionTreeBase

from ._classifier import Classifier

if TYPE_CHECKING:
    from sklearn.base import ClassifierMixin


class DecisionTreeClassifier(Classifier, _DecisionTreeBase):
    """
    Decision tree classification.

    Parameters
    ----------
    max_depth:
        The maximum depth of each tree. If None, the depth is not limited. Has to be greater than 0.
    min_sample_count_in_leaves:
        The minimum number of samples that must remain in the leaves of each tree. Has to be greater than 0.

    Raises
    ------
    OutOfBoundsError
        If `max_depth` is less than 1.
    OutOfBoundsError
        If `min_sample_count_in_leaves` is less than 1.
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(
        self,
        *,
        max_depth: int | None = None,
        min_sample_count_in_leaves: int = 1,
    ) -> None:
        # Initialize superclasses
        Classifier.__init__(self)
        _DecisionTreeBase.__init__(
            self,
            max_depth=max_depth,
            min_sample_count_in_leaves=min_sample_count_in_leaves,
        )

    def __hash__(self) -> int:
        return _structural_hash(
            Classifier.__hash__(self),
            _DecisionTreeBase.__hash__(self),
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Template methods
    # ------------------------------------------------------------------------------------------------------------------

    def _clone(self) -> DecisionTreeClassifier:
        return DecisionTreeClassifier(
            max_depth=self._max_depth,
            min_sample_count_in_leaves=self._min_sample_count_in_leaves,
        )

    def _get_sklearn_model(self) -> ClassifierMixin:
        from sklearn.tree import DecisionTreeClassifier as SklearnDecisionTreeClassifier

        return SklearnDecisionTreeClassifier(
            max_depth=self._max_depth,
            min_samples_leaf=self._min_sample_count_in_leaves,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Plot
    # ------------------------------------------------------------------------------------------------------------------

    def plot(self) -> Image:
        """
        Get the image of the decision tree.

        Returns
        -------
        plot:
            The decision tree figure as an image.

        Raises
        ------
        ModelNotFittedError:
            If model is not fitted.
        """
        if not self.is_fitted:
            raise ModelNotFittedError

        from io import BytesIO

        import matplotlib.pyplot as plt
        from sklearn.tree import plot_tree

        plot_tree(self._wrapped_model)

        # save plot fig bytes in buffer
        with BytesIO() as buffer:
            plt.savefig(buffer)
            image = buffer.getvalue()

        # prevent forced plot from sklearn showing
        plt.close()

        return Image.from_bytes(image)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, overload

from safeds._config import _get_device, _init_default_device
from safeds._config._polars import _get_polars_config
from safeds._utils import _structural_hash
from safeds._utils._random import _get_random_seed
from safeds._validation import _check_bounds, _check_columns_exist, _ClosedBound, _normalize_and_check_file_path
from safeds._validation._check_columns_dont_exist import _check_columns_dont_exist
from safeds.data.tabular.plotting import TablePlotter
from safeds.data.tabular.typing._polars_data_type import _PolarsDataType
from safeds.data.tabular.typing._polars_schema import _PolarsSchema
from safeds.exceptions import (
    ColumnLengthMismatchError,
    DuplicateColumnError,
)

from ._column import Column
from ._lazy_cell import _LazyCell
from ._lazy_vectorized_row import _LazyVectorizedRow

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence
    from pathlib import Path

    import polars as pl
    import torch
    from torch import Tensor
    from torch.utils.data import DataLoader, Dataset

    from safeds.data.labeled.containers import TabularDataset, TimeSeriesDataset
    from safeds.data.tabular.transformation import (
        InvertibleTableTransformer,
        TableTransformer,
    )
    from safeds.data.tabular.typing import DataType, Schema

    from ._cell import Cell
    from ._row import Row


class Table:
    """
    A two-dimensional collection of data. It can either be seen as a list of rows or as a list of columns.

    To create a `Table` call the constructor or use one of the following static methods:

    | Method                                                                             | Description                            |
    | ---------------------------------------------------------------------------------- | -------------------------------------- |
    | [from_csv_file][safeds.data.tabular.containers._table.Table.from_csv_file]         | Create a table from a CSV file.        |
    | [from_json_file][safeds.data.tabular.containers._table.Table.from_json_file]       | Create a table from a JSON file.       |
    | [from_parquet_file][safeds.data.tabular.containers._table.Table.from_parquet_file] | Create a table from a Parquet file.    |
    | [from_columns][safeds.data.tabular.containers._table.Table.from_columns]           | Create a table from a list of columns. |
    | [from_dict][safeds.data.tabular.containers._table.Table.from_dict]                 | Create a table from a dictionary.      |

    Parameters
    ----------
    data:
        The data of the table. If None, an empty table is created.

    Raises
    ------
    ValueError
        If columns have different lengths.

    Examples
    --------
    >>> from safeds.data.tabular.containers import Table
    >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Import
    # ------------------------------------------------------------------------------------------------------------------

    @staticmethod
    def from_columns(columns: Column | list[Column]) -> Table:
        """
        Create a table from a list of columns.

        Parameters
        ----------
        columns:
            The columns.

        Returns
        -------
        table:
            The created table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column, Table
        >>> a = Column("a", [1, 2, 3])
        >>> b = Column("b", [4, 5, 6])
        >>> Table.from_columns([a, b])
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   2 |   5 |
        |   3 |   6 |
        +-----+-----+
        """
        import polars as pl
        from polars.exceptions import DuplicateError, ShapeError

        if isinstance(columns, Column):
            columns = [columns]

        try:
            return Table._from_polars_lazy_frame(
                pl.LazyFrame([column._series for column in columns]),
            )
        except DuplicateError:
            raise DuplicateColumnError("") from None  # TODO: message
        except ShapeError:
            raise ColumnLengthMismatchError("") from None  # TODO: message

    @staticmethod
    def from_csv_file(path: str | Path, *, separator: str = ",") -> Table:
        """
        Create a table from a CSV file.

        Parameters
        ----------
        path:
            The path to the CSV file. If the file extension is omitted, it is assumed to be ".csv".
        separator:
            The separator between the values in the CSV file.

        Returns
        -------
        table:
            The created table.

        Raises
        ------
        FileNotFoundError
            If no file exists at the given path.
        ValueError
            If the path has an extension that is not ".csv".

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> Table.from_csv_file("./src/resources/from_csv_file.csv")
        +-----+-----+-----+
        |   a |   b |   c |
        | --- | --- | --- |
        | i64 | i64 | i64 |
        +=================+
        |   1 |   2 |   1 |
        |   0 |   0 |   7 |
        +-----+-----+-----+
        """
        import polars as pl

        path = _normalize_and_check_file_path(path, ".csv", [".csv"], check_if_file_exists=True)

        return Table._from_polars_lazy_frame(pl.scan_csv(path, separator=separator, raise_if_empty=False))

    @staticmethod
    def from_dict(data: dict[str, list[Any]]) -> Table:
        """
        Create a table from a dictionary that maps column names to column values.

        Parameters
        ----------
        data:
            The data.

        Returns
        -------
        table:
            The generated table.

        Raises
        ------
        ValueError
            If columns have different lengths.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> data = {'a': [1, 2, 3], 'b': [4, 5, 6]}
        >>> Table.from_dict(data)
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   2 |   5 |
        |   3 |   6 |
        +-----+-----+
        """
        return Table(data)

    @staticmethod
    def from_json_file(path: str | Path) -> Table:
        """
        Create a table from a JSON file.

        Parameters
        ----------
        path:
            The path to the JSON file. If the file extension is omitted, it is assumed to be ".json".

        Returns
        -------
        table:
            The created table.

        Raises
        ------
        FileNotFoundError
            If no file exists at the given path.
        ValueError
            If the path has an extension that is not ".json".

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> Table.from_json_file("./src/resources/from_json_file.json")
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   2 |   5 |
        |   3 |   6 |
        +-----+-----+
        """
        import polars as pl

        path = _normalize_and_check_file_path(path, ".json", [".json"], check_if_file_exists=True)

        try:
            return Table._from_polars_data_frame(pl.read_json(path))
        except pl.PolarsPanicError:
            # Can happen if the JSON file is empty (https://github.com/pola-rs/polars/issues/10234)
            return Table()

    @staticmethod
    def from_parquet_file(path: str | Path) -> Table:
        """
        Create a table from a Parquet file.

        Parameters
        ----------
        path:
            The path to the Parquet file. If the file extension is omitted, it is assumed to be ".parquet".

        Returns
        -------
        table:
            The created table.

        Raises
        ------
        FileNotFoundError
            If no file exists at the given path.
        ValueError
            If the path has an extension that is not ".parquet".

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> Table.from_parquet_file("./src/resources/from_parquet_file.parquet")
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   2 |   5 |
        |   3 |   6 |
        +-----+-----+
        """
        import polars as pl

        path = _normalize_and_check_file_path(path, ".parquet", [".parquet"], check_if_file_exists=True)
        return Table._from_polars_lazy_frame(pl.scan_parquet(path))

    @staticmethod
    def _from_polars_data_frame(data: pl.DataFrame) -> Table:
        result = object.__new__(Table)
        result._lazy_frame = data.lazy()
        result.__data_frame_cache = data
        return result

    @staticmethod
    def _from_polars_lazy_frame(data: pl.LazyFrame) -> Table:
        result = object.__new__(Table)
        result._lazy_frame = data
        result.__data_frame_cache = None
        return result

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, data: Mapping[str, Sequence[Any]] | None = None) -> None:
        import polars as pl

        if data is None:
            data = {}

        # Validation
        expected_length: int | None = None
        for column_values in data.values():
            if expected_length is None:
                expected_length = len(column_values)
            elif len(column_values) != expected_length:
                raise ColumnLengthMismatchError(
                    "\n".join(f"{column_name}: {len(column_values)}" for column_name, column_values in data.items()),
                )

        # Implementation
        self._lazy_frame: pl.LazyFrame = pl.LazyFrame(data)
        self.__data_frame_cache: pl.DataFrame | None = None  # Scramble the name to prevent access from outside

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Table):
            return NotImplemented
        if self is other:
            return True

        return self._data_frame.equals(other._data_frame)

    def __hash__(self) -> int:
        return _structural_hash(self.schema, self.row_count)

    def __repr__(self) -> str:
        with _get_polars_config():
            return self._data_frame.__repr__()

    def __sizeof__(self) -> int:
        return self._data_frame.estimated_size()

    def __str__(self) -> str:
        with _get_polars_config():
            return self._data_frame.__str__()

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def _data_frame(self) -> pl.DataFrame:
        import polars as pl

        if self.__data_frame_cache is None:
            try:
                self.__data_frame_cache = self._lazy_frame.collect()
            except (pl.NoDataError, pl.PolarsPanicError):
                # Can happen for some operations on empty tables (e.g. https://github.com/pola-rs/polars/issues/16202)
                return pl.DataFrame()

        return self.__data_frame_cache

    @property
    def column_names(self) -> list[str]:
        """
        The names of the columns in the table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.column_names
        ['a', 'b']
        """
        return self.schema.column_names

    @property
    def column_count(self) -> int:
        """
        The number of columns in the table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.column_count
        2
        """
        import polars as pl

        try:
            return self._lazy_frame.width
        except (pl.NoDataError, pl.PolarsPanicError):
            # Can happen for some operations on empty tables (e.g. https://github.com/pola-rs/polars/issues/16202)
            return 0

    @property
    def row_count(self) -> int:
        """
        The number of rows in the table.

        **Note:** This operation must fully load the data into memory, which can be expensive.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.row_count
        3
        """
        return self._data_frame.height

    @property
    def plot(self) -> TablePlotter:
        """The plotter for the table."""
        return TablePlotter(self)

    @property
    def schema(self) -> Schema:
        """The schema of the table."""
        import polars as pl

        try:
            return _PolarsSchema(self._lazy_frame.schema)
        except (pl.NoDataError, pl.PolarsPanicError):
            # Can happen for some operations on empty tables (e.g. https://github.com/pola-rs/polars/issues/16202)
            return _PolarsSchema({})

    # ------------------------------------------------------------------------------------------------------------------
    # Column operations
    # ------------------------------------------------------------------------------------------------------------------

    def add_columns(
        self,
        columns: Column | list[Column],
    ) -> Table:
        """
        Return a new table with additional columns.

        **Notes:**

        - The original table is not modified.
        - This operation must fully load the data into memory, which can be expensive.

        Parameters
        ----------
        columns:
            The columns to add.

        Returns
        -------
        new_table:
            The table with the additional columns.

        Raises
        ------
        ValueError
            If a column name already exists.
        ValueError
            If the columns have incompatible lengths.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column, Table
        >>> table = Table({"a": [1, 2, 3]})
        >>> new_column = Column("b", [4, 5, 6])
        >>> table.add_columns(new_column)
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   2 |   5 |
        |   3 |   6 |
        +-----+-----+
        """
        import polars as pl

        if isinstance(columns, Column):
            columns = [columns]

        if len(columns) == 0:
            return self

        try:
            return Table._from_polars_data_frame(
                self._data_frame.hstack([column._series for column in columns]),
            )
        except pl.DuplicateError:
            # polars already validates this, so we don't need to do it again upfront (performance)
            _check_columns_dont_exist(self, [column.name for column in columns])
            return Table()  # pragma: no cover

    def add_computed_column(
        self,
        name: str,
        computer: Callable[[Row], Cell],
    ) -> Table:
        """
        Return a new table with an additional computed column.

        **Note:** The original table is not modified.

        Parameters
        ----------
        name:
            The name of the new column.
        computer:
            The function that computes the values of the new column.

        Returns
        -------
        new_table:
            The table with the computed column.

        Raises
        ------
        ValueError
            If the column name already exists.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.add_computed_column("c", lambda row: row.get_value("a") + row.get_value("b"))
        +-----+-----+-----+
        |   a |   b |   c |
        | --- | --- | --- |
        | i64 | i64 | i64 |
        +=================+
        |   1 |   4 |   5 |
        |   2 |   5 |   7 |
        |   3 |   6 |   9 |
        +-----+-----+-----+
        """
        if self.has_column(name):
            raise DuplicateColumnError(name)

        computed_column = computer(_LazyVectorizedRow(self))

        return self._from_polars_lazy_frame(
            self._lazy_frame.with_columns(computed_column._polars_expression.alias(name)),
        )

    def get_column(self, name: str) -> Column:
        """
        Get a column from the table.

        Parameters
        ----------
        name:
            The name of the column.

        Returns
        -------
        column:
            The column.

        Raises
        ------
        ColumnNotFoundError
            If the column does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.get_column("a")
        +-----+
        |   a |
        | --- |
        | i64 |
        +=====+
        |   1 |
        |   2 |
        |   3 |
        +-----+
        """
        _check_columns_exist(self, name)
        return Column._from_polars_series(
            self._lazy_frame.select(name).collect().get_column(name),
        )

    def get_column_type(self, name: str) -> DataType:
        """
        Get the data type of a column.

        Parameters
        ----------
        name:
            The name of the column.

        Returns
        -------
        type:
            The data type of the column.

        Raises
        ------
        ColumnNotFoundError
            If the column does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.get_column_type("a")
        Int64
        """
        _check_columns_exist(self, name)
        return _PolarsDataType(self._lazy_frame.schema[name])

    def has_column(self, name: str) -> bool:
        """
        Check if the table has a column with a specific name.

        Parameters
        ----------
        name:
            The name of the column.

        Returns
        -------
        has_column:
            Whether the table has a column with the specified name.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.has_column("a")
        True
        """
        return name in self.column_names

    def remove_columns(
        self,
        names: str | list[str],
        /,
        *,
        ignore_unknown_names: bool = False,
    ) -> Table:
        """
        Return a new table without the specified columns.

        **Notes:**

        - The original table is not modified.

        Parameters
        ----------
        names:
            The names of the columns to remove.
        ignore_unknown_names:
            If set to True, columns that are not present in the table will be ignored.
            If set to False, an error will be raised if any of the specified columns do not exist.

        Returns
        -------
        new_table:
            The table with the columns removed.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.remove_columns("a")
        +-----+
        |   b |
        | --- |
        | i64 |
        +=====+
        |   4 |
        |   5 |
        |   6 |
        +-----+

        >>> table.remove_columns(["c"], ignore_unknown_names=True)
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   2 |   5 |
        |   3 |   6 |
        +-----+-----+
        """
        if isinstance(names, str):
            names = [names]

        if not ignore_unknown_names:
            _check_columns_exist(self, names)

        return Table._from_polars_lazy_frame(
            self._lazy_frame.drop(names),
        )

    def remove_columns_except(
        self,
        names: str | list[str],
        /,
    ) -> Table:
        """
        Return a new table with only the specified columns.

        Parameters
        ----------
        names:
            The names of the columns to keep.

        Returns
        -------
        new_table:
            The table with only the specified columns.

        Raises
        ------
        ColumnNotFoundError
            If a column does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.remove_columns_except("a")
        +-----+
        |   a |
        | --- |
        | i64 |
        +=====+
        |   1 |
        |   2 |
        |   3 |
        +-----+
        """
        if isinstance(names, str):
            names = [names]

        _check_columns_exist(self, names)

        return Table._from_polars_lazy_frame(
            self._lazy_frame.select(names),
        )

    def remove_columns_with_missing_values(self) -> Table:
        """
        Return a new table without columns that contain missing values.

        **Notes:**

        - The original table is not modified.
        - This operation must fully load the data into memory, which can be expensive.

        Returns
        -------
        new_table:
            The table without columns containing missing values.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, None]})
        >>> table.remove_columns_with_missing_values()
        +-----+
        |   a |
        | --- |
        | i64 |
        +=====+
        |   1 |
        |   2 |
        |   3 |
        +-----+
        """
        import polars as pl

        return Table._from_polars_lazy_frame(
            pl.LazyFrame(
                [series for series in self._data_frame.get_columns() if series.null_count() == 0],
            ),
        )

    def remove_non_numeric_columns(self) -> Table:
        """
        Return a new table without non-numeric columns.

        **Note:** The original table is not modified.

        Returns
        -------
        new_table:
            The table without non-numeric columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": ["4", "5", "6"]})
        >>> table.remove_non_numeric_columns()
        +-----+
        |   a |
        | --- |
        | i64 |
        +=====+
        |   1 |
        |   2 |
        |   3 |
        +-----+
        """
        import polars.selectors as cs

        return Table._from_polars_lazy_frame(
            self._lazy_frame.select(cs.numeric()),
        )

    def rename_column(self, old_name: str, new_name: str) -> Table:
        """
        Return a new table with a column renamed.

        **Note:** The original table is not modified.

        Parameters
        ----------
        old_name:
            The name of the column to rename.
        new_name:
            The new name of the column.

        Returns
        -------
        new_table:
            The table with the column renamed.

        Raises
        ------
        ColumnNotFoundError
            If no column with the old name exists.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.rename_column("a", "c")
        +-----+-----+
        |   c |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   2 |   5 |
        |   3 |   6 |
        +-----+-----+
        """
        _check_columns_exist(self, old_name)
        _check_columns_dont_exist(self, new_name, old_name=old_name)

        return Table._from_polars_lazy_frame(
            self._lazy_frame.rename({old_name: new_name}),
        )

    def replace_column(
        self,
        old_name: str,
        new_columns: Column | list[Column] | Table,
    ) -> Table:
        """
        Return a new table with a column replaced by zero or more columns.

        **Note:** The original table is not modified.

        Parameters
        ----------
        old_name:
            The name of the column to replace.
        new_columns:
            The new columns.

        Returns
        -------
        new_table:
            The table with the column replaced.

        Raises
        ------
        ColumnNotFoundError
            If no column with the old name exists.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Column, Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.replace_column("a", [])
        +-----+
        |   b |
        | --- |
        | i64 |
        +=====+
        |   4 |
        |   5 |
        |   6 |
        +-----+

        >>> column1 = Column("c", [7, 8, 9])
        >>> table.replace_column("a", column1)
        +-----+-----+
        |   c |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   7 |   4 |
        |   8 |   5 |
        |   9 |   6 |
        +-----+-----+

        >>> column2 = Column("d", [10, 11, 12])
        >>> table.replace_column("a", [column1, column2])
        +-----+-----+-----+
        |   c |   d |   b |
        | --- | --- | --- |
        | i64 | i64 | i64 |
        +=================+
        |   7 |  10 |   4 |
        |   8 |  11 |   5 |
        |   9 |  12 |   6 |
        +-----+-----+-----+
        """
        if isinstance(new_columns, Column):
            new_columns = [new_columns]
        elif isinstance(new_columns, Table):
            new_columns = new_columns.to_columns()

        _check_columns_exist(self, old_name)
        _check_columns_dont_exist(self, [column.name for column in new_columns], old_name=old_name)

        if len(new_columns) == 0:
            return self.remove_columns(old_name, ignore_unknown_names=True)

        if len(new_columns) == 1:
            new_column = new_columns[0]
            return Table._from_polars_lazy_frame(
                self._lazy_frame.with_columns(new_column._series.alias(old_name)).rename({old_name: new_column.name}),
            )

        import polars as pl

        index = self.column_names.index(old_name)

        return Table._from_polars_lazy_frame(
            self._lazy_frame.select(
                *[pl.col(name) for name in self.column_names[:index]],
                *[column._series for column in new_columns],
                *[pl.col(name) for name in self.column_names[index + 1 :]],
            ),
        )

    def transform_column(
        self,
        name: str,
        transformer: Callable[[Cell], Cell],
    ) -> Table:
        """
        Return a new table with a column transformed.

        **Note:** The original table is not modified.

        Parameters
        ----------
        name:
            The name of the column to transform.

        transformer:
            The function that transforms the column.

        Returns
        -------
        new_table:
            The table with the transformed column.

        Raises
        ------
        ColumnNotFoundError
            If no column with the specified name exists.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.transform_column("a", lambda cell: cell + 1)
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   2 |   4 |
        |   3 |   5 |
        |   4 |   6 |
        +-----+-----+
        """
        _check_columns_exist(self, name)

        import polars as pl

        expression = transformer(_LazyCell(pl.col(name)))

        return Table._from_polars_lazy_frame(
            self._lazy_frame.with_columns(expression._polars_expression),
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Row operations
    # ------------------------------------------------------------------------------------------------------------------

    @overload
    def count_row_if(
        self,
        predicate: Callable[[Row], Cell[bool | None]],
        *,
        ignore_unknown: Literal[True] = ...,
    ) -> int: ...

    @overload
    def count_row_if(
        self,
        predicate: Callable[[Row], Cell[bool | None]],
        *,
        ignore_unknown: bool,
    ) -> int | None: ...

    def count_row_if(
        self,
        predicate: Callable[[Row], Cell[bool | None]],
        *,
        ignore_unknown: bool = True,
    ) -> int | None:
        """
        Return how many rows in the table satisfy the predicate.

        The predicate can return one of three results:

        * True, if the row satisfies the predicate.
        * False, if the row does not satisfy the predicate.
        * None, if the truthiness of the predicate is unknown, e.g. due to missing values.

        By default, cases where the truthiness of the predicate is unknown are ignored and this method returns how
        often the predicate returns True.

        You can instead enable Kleene logic by setting `ignore_unknown=False`. In this case, this method returns None if
        the predicate returns None at least once. Otherwise, it still returns how often the predicate returns True.

        Parameters
        ----------
        predicate:
            The predicate to apply to each row.
        ignore_unknown:
            Whether to ignore cases where the truthiness of the predicate is unknown.

        Returns
        -------
        count:
            The number of rows in the table that satisfy the predicate.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"col1": [1, 2, 3], "col2": [1, 3, 3]})
        >>> table.count_row_if(lambda row: row["col1"] == row["col2"])
        2

        >>> table.count_row_if(lambda row: row["col1"] > row["col2"])
        0
        """
        expression = predicate(_LazyVectorizedRow(self))._polars_expression
        series = self._lazy_frame.select(expression.alias("count")).collect().get_column("count")

        if ignore_unknown or series.null_count() == 0:
            return series.sum()
        else:
            return None

    # TODO: Rethink group_rows/group_rows_by_column. They should not return a dict.

    def remove_duplicate_rows(self) -> Table:
        """
        Return a new table without duplicate rows.

        **Note:** The original table is not modified.

        Returns
        -------
        new_table:
            The table without duplicate rows.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 2], "b": [4, 5, 5]})
        >>> table.remove_duplicate_rows()
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   2 |   5 |
        +-----+-----+
        """
        return Table._from_polars_lazy_frame(
            self._lazy_frame.unique(maintain_order=True),
        )

    def remove_rows(
        self,
        query: Callable[[Row], Cell[bool]],
    ) -> Table:
        """
        Return a new table without rows that satisfy a condition.

        **Note:** The original table is not modified.

        Parameters
        ----------
        query:
            The function that determines which rows to remove.

        Returns
        -------
        new_table:
            The table without the specified rows.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.remove_rows(lambda row: row.get_value("a") == 2)
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   3 |   6 |
        +-----+-----+
        """
        mask = query(_LazyVectorizedRow(self))

        return Table._from_polars_lazy_frame(
            self._lazy_frame.filter(~mask._polars_expression),
        )

    def remove_rows_by_column(
        self,
        name: str,
        query: Callable[[Cell], Cell[bool]],
    ) -> Table:
        """
        Return a new table without rows that satisfy a condition on a specific column.

        **Note:** The original table is not modified.

        Parameters
        ----------
        name:
            The name of the column.
        query:
            The function that determines which rows to remove.

        Returns
        -------
        new_table:
            The table without the specified rows.

        Raises
        ------
        ColumnNotFoundError
            If the column does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.remove_rows_by_column("a", lambda cell: cell == 2)
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   3 |   6 |
        +-----+-----+
        """
        _check_columns_exist(self, name)

        import polars as pl

        mask = query(_LazyCell(pl.col(name)))

        return Table._from_polars_lazy_frame(
            self._lazy_frame.filter(~mask._polars_expression),
        )

    def remove_rows_with_missing_values(
        self,
        column_names: list[str] | None = None,
    ) -> Table:
        """
        Return a new table without rows containing missing values in the specified columns.

        **Note:** The original table is not modified.

        Parameters
        ----------
        column_names:
            Names of the columns to consider. If None, all columns are considered.

        Returns
        -------
        new_table:
            The table without rows containing missing values in the specified columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, None, 3], "b": [4, 5, None]})
        >>> table.remove_rows_with_missing_values()
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        +-----+-----+
        """
        return Table._from_polars_lazy_frame(
            self._lazy_frame.drop_nulls(subset=column_names),
        )

    def remove_rows_with_outliers(
        self,
        column_names: list[str] | None = None,
        *,
        z_score_threshold: float = 3,
    ) -> Table:
        """
        Return a new table without rows containing outliers in the specified columns.

        Whether a data point is an outlier in a column is determined by its z-score. The z-score the distance of the
        data point from the mean of the column divided by the standard deviation of the column. If the z-score is
        greater than the given threshold, the data point is considered an outlier. Missing values are ignored during the
        calculation of the z-score.

        The z-score is only defined for numeric columns. Non-numeric columns are ignored, even if they are specified in
        `column_names`.

        **Notes:**

        - The original table is not modified.
        - This operation must fully load the data into memory, which can be expensive.

        Parameters
        ----------
        column_names:
            Names of the columns to consider. If None, all numeric columns are considered.
        z_score_threshold:
            The z-score threshold for detecting outliers.

        Returns
        -------
        new_table:
            The table without rows containing outliers in the specified columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table(
        ...     {
        ...         "a": [1, 2, 3, 4, 5, 6, 1000, None],
        ...         "b": [1, 2, 3, 4, 5, 6,    7,    8],
        ...     }
        ... )
        >>> table.remove_rows_with_outliers(z_score_threshold=2)
        +------+-----+
        |    a |   b |
        |  --- | --- |
        |  i64 | i64 |
        +============+
        |    1 |   1 |
        |    2 |   2 |
        |    3 |   3 |
        |    4 |   4 |
        |    5 |   5 |
        |    6 |   6 |
        | null |   8 |
        +------+-----+
        """
        if self.row_count == 0:
            return self  # polars raises a ComputeError for tables without rows
        if column_names is None:
            column_names = self.column_names

        import polars as pl
        import polars.selectors as cs

        non_outlier_mask = pl.all_horizontal(
            self._data_frame.select(cs.numeric() & cs.by_name(column_names)).select(
                pl.all().is_null() | (((pl.all() - pl.all().mean()) / pl.all().std()).abs() <= z_score_threshold),
            ),
        )

        return Table._from_polars_lazy_frame(
            self._lazy_frame.filter(non_outlier_mask),
        )

    def shuffle_rows(self) -> Table:
        """
        Return a new table with the rows shuffled.

        **Note:** The original table is not modified.

        Returns
        -------
        new_table:
            The table with the rows shuffled.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.shuffle_rows()
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   3 |   6 |
        |   2 |   5 |
        |   1 |   4 |
        +-----+-----+
        """
        return Table._from_polars_data_frame(
            self._data_frame.sample(
                fraction=1,
                shuffle=True,
                seed=_get_random_seed(),
            ),
        )

    def slice_rows(self, start: int = 0, length: int | None = None) -> Table:
        """
        Return a new table with a slice of rows.

        **Note:** The original table is not modified.

        Parameters
        ----------
        start:
            The start index of the slice.
        length:
            The length of the slice. If None, the slice contains all rows starting from `start`. Must greater than or
            equal to 0.

        Returns
        -------
        new_table:
            The table with the slice of rows.

        Raises
        ------
        OutOfBoundsError
            If length is less than 0.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.slice_rows(start=1)
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   2 |   5 |
        |   3 |   6 |
        +-----+-----+

        >>> table.slice_rows(start=1, length=1)
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   2 |   5 |
        +-----+-----+
        """
        _check_bounds("length", length, lower_bound=_ClosedBound(0))

        return Table._from_polars_lazy_frame(
            self._lazy_frame.slice(start, length),
        )

    def sort_rows(
        self,
        key_selector: Callable[[Row], Cell],
        *,
        descending: bool = False,
    ) -> Table:
        """
        Return a new table with the rows sorted.

        **Note:** The original table is not modified.

        Parameters
        ----------
        key_selector:
            The function that selects the key to sort by.
        descending:
            Whether to sort in descending order.

        Returns
        -------
        new_table:
            The table with the rows sorted.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [2, 1, 3], "b": [1, 1, 2]})
        >>> table.sort_rows(lambda row: row.get_value("a") - row.get_value("b"))
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   1 |
        |   2 |   1 |
        |   3 |   2 |
        +-----+-----+
        """
        if self.row_count == 0:
            return self

        key = key_selector(_LazyVectorizedRow(self))

        return Table._from_polars_lazy_frame(
            self._lazy_frame.sort(
                key._polars_expression,
                descending=descending,
                maintain_order=True,
            ),
        )

    def sort_rows_by_column(
        self,
        name: str,
        *,
        descending: bool = False,
    ) -> Table:
        """
        Return a new table with the rows sorted by a specific column.

        **Note:** The original table is not modified.

        Parameters
        ----------
        name:
            The name of the column to sort by.
        descending:
            Whether to sort in descending order.

        Returns
        -------
        new_table:
            The table with the rows sorted by the specified column.

        Raises
        ------
        ColumnNotFoundError
            If the column does not exist.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [2, 1, 3], "b": [1, 1, 2]})
        >>> table.sort_rows_by_column("a")
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   1 |
        |   2 |   1 |
        |   3 |   2 |
        +-----+-----+
        """
        _check_columns_exist(self, name)

        return Table._from_polars_lazy_frame(
            self._lazy_frame.sort(
                name,
                descending=descending,
                maintain_order=True,
            ),
        )

    def split_rows(
        self,
        percentage_in_first: float,
        *,
        shuffle: bool = True,
    ) -> tuple[Table, Table]:
        """
        Create two tables by splitting the rows of the current table.

        The first table contains a percentage of the rows specified by `percentage_in_first`, and the second table
        contains the remaining rows.

        **Notes:**

        - The original table is not modified.
        - By default, the rows are shuffled before splitting. You can disable this by setting `shuffle` to False.

        Parameters
        ----------
        percentage_in_first:
            The percentage of rows to include in the first table. Must be between 0 and 1.
        shuffle:
            Whether to shuffle the rows before splitting.

        Returns
        -------
        first_table:
            The first table.
        second_table:
            The second table.

        Raises
        ------
        OutOfBoundsError
            If `percentage_in_first` is not between 0 and 1.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3, 4, 5], "b": [6, 7, 8, 9, 10]})
        >>> first_table, second_table = table.split_rows(0.6)
        >>> first_table
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   6 |
        |   4 |   9 |
        |   3 |   8 |
        +-----+-----+
        >>> second_table
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   5 |  10 |
        |   2 |   7 |
        +-----+-----+
        """
        _check_bounds(
            "percentage_in_first",
            percentage_in_first,
            lower_bound=_ClosedBound(0),
            upper_bound=_ClosedBound(1),
        )

        input_table = self.shuffle_rows() if shuffle else self
        row_count_in_first = round(percentage_in_first * input_table.row_count)

        return (
            input_table.slice_rows(length=row_count_in_first),
            input_table.slice_rows(start=row_count_in_first),
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Table operations
    # ------------------------------------------------------------------------------------------------------------------

    def add_table_as_columns(self, other: Table) -> Table:
        """
        Return a new table with the columns of another table added.

        **Notes:**

        - The original tables are not modified.
        - This operation must fully load the data into memory, which can be expensive.

        Parameters
        ----------
        other:
            The table to add as columns.

        Returns
        -------
        new_table:
            The table with the columns added.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table1 = Table({"a": [1, 2, 3]})
        >>> table2 = Table({"b": [4, 5, 6]})
        >>> table1.add_table_as_columns(table2)
        +-----+-----+
        |   a |   b |
        | --- | --- |
        | i64 | i64 |
        +===========+
        |   1 |   4 |
        |   2 |   5 |
        |   3 |   6 |
        +-----+-----+
        """
        # TODO: raises?

        return Table._from_polars_data_frame(
            self._data_frame.hstack(other._data_frame),
        )

    def add_table_as_rows(self, other: Table) -> Table:
        """
        Return a new table with the rows of another table added.

        **Notes:**

        - The original tables are not modified.
        - This operation must fully load the data into memory, which can be expensive.

        Parameters
        ----------
        other:
            The table to add as rows.

        Returns
        -------
        new_table:
            The table with the rows added.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table1 = Table({"a": [1, 2, 3]})
        >>> table2 = Table({"a": [4, 5, 6]})
        >>> table1.add_table_as_rows(table2)
        +-----+
        |   a |
        | --- |
        | i64 |
        +=====+
        |   1 |
        |   2 |
        |   3 |
        |   4 |
        |   5 |
        |   6 |
        +-----+
        """
        # TODO: raises?

        return Table._from_polars_data_frame(
            self._data_frame.vstack(other._data_frame),
        )

    def inverse_transform_table(self, fitted_transformer: InvertibleTableTransformer) -> Table:
        """
        Return a new table inverse-transformed by a **fitted, invertible** transformer.

        **Notes:**

        - The original table is not modified.
        - Depending on the transformer, this operation might fully load the data into memory, which can be expensive.

        Parameters
        ----------
        fitted_transformer:
            The fitted, invertible transformer to apply.

        Returns
        -------
        new_table:
            The inverse-transformed table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> from safeds.data.tabular.transformation import RangeScaler
        >>> table = Table({"a": [1, 2, 3]})
        >>> transformer, transformed_table = RangeScaler(min_=0, max_=1, column_names="a").fit_and_transform(table)
        >>> transformed_table.inverse_transform_table(transformer)
        +---------+
        |       a |
        |     --- |
        |     f64 |
        +=========+
        | 1.00000 |
        | 2.00000 |
        | 3.00000 |
        +---------+
        """
        return fitted_transformer.inverse_transform(self)

    def join(
        self,
        right_table: Table,
        left_names: str | list[str],
        right_names: str | list[str],
        *,
        mode: Literal["inner", "left", "outer"] = "inner",
    ) -> Table:
        """
        Join a table with the current table and return the result.

        Parameters
        ----------
        right_table:
            The other table which is to be joined to the current table.
        left_names:
            Name or list of names of columns from the current table on which to join right_table.
        right_names:
            Name or list of names of columns from right_table on which to join the current table.
        mode:
            Specify which type of join you want to use. Options include 'inner', 'outer', 'left', 'right'.

        Returns
        -------
        new_table:
            The table with the joined table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table1 = Table({"a": [1, 2], "b": [3, 4]})
        >>> table2 = Table({"d": [1, 5], "e": [5, 6]})
        >>> table1.join(table2, "a", "d", mode="left")
        +-----+-----+------+
        |   a |   b |    e |
        | --- | --- |  --- |
        | i64 | i64 |  i64 |
        +==================+
        |   1 |   3 |    5 |
        |   2 |   4 | null |
        +-----+-----+------+
        """
        # Validation
        _check_columns_exist(self, left_names)
        _check_columns_exist(right_table, right_names)

        if len(left_names) != len(right_names):
            raise ValueError("The number of columns to join on must be the same in both tables.")

        # Implementation
        return self._from_polars_lazy_frame(
            self._lazy_frame.join(
                right_table._lazy_frame,
                left_on=left_names,
                right_on=right_names,
                how=mode,
            ),
        )

    def transform_table(self, fitted_transformer: TableTransformer) -> Table:
        """
        Return a new table transformed by a **fitted** transformer.

        **Notes:**

        - The original table is not modified.
        - Depending on the transformer, this operation might fully load the data into memory, which can be expensive.

        Parameters
        ----------
        fitted_transformer:
            The fitted transformer to apply.

        Returns
        -------
        new_table:
            The transformed table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> from safeds.data.tabular.transformation import RangeScaler
        >>> table = Table({"a": [1, 2, 3]})
        >>> transformer = RangeScaler(min_=0, max_=1, column_names="a").fit(table)
        >>> table.transform_table(transformer)
        +---------+
        |       a |
        |     --- |
        |     f64 |
        +=========+
        | 0.00000 |
        | 0.50000 |
        | 1.00000 |
        +---------+
        """
        return fitted_transformer.transform(self)

    # ------------------------------------------------------------------------------------------------------------------
    # Statistics
    # ------------------------------------------------------------------------------------------------------------------

    def summarize_statistics(self) -> Table:
        """
        Return a table with important statistics about this table.

        Returns
        -------
        statistics:
            The table with statistics.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 3]})
        >>> table.summarize_statistics()
        +----------------------+---------+
        | metric               |       a |
        | ---                  |     --- |
        | str                  |     f64 |
        +================================+
        | min                  | 1.00000 |
        | max                  | 3.00000 |
        | mean                 | 2.00000 |
        | median               | 2.00000 |
        | standard deviation   | 1.41421 |
        | distinct value count | 2.00000 |
        | idness               | 1.00000 |
        | missing value ratio  | 0.00000 |
        | stability            | 0.50000 |
        +----------------------+---------+
        """
        if self.column_count == 0:
            return Table()

        head = self.get_column(self.column_names[0]).summarize_statistics()
        tail = [self.get_column(name).summarize_statistics().get_column(name)._series for name in self.column_names[1:]]

        return Table._from_polars_data_frame(
            head._lazy_frame.collect().hstack(tail, in_place=True),
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Export
    # ------------------------------------------------------------------------------------------------------------------

    def to_columns(self) -> list[Column]:
        """
        Return the data of the table as a list of columns.

        Returns
        -------
        columns:
            List of columns.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> columns = table.to_columns()
        """
        return [Column._from_polars_series(column) for column in self._data_frame.get_columns()]

    def to_csv_file(self, path: str | Path) -> None:
        """
        Write the table to a CSV file.

        If the file and/or the parent directories do not exist, they will be created. If the file exists already, it
        will be overwritten.

        Parameters
        ----------
        path:
            The path to the CSV file. If the file extension is omitted, it is assumed to be ".csv".

        Raises
        ------
        ValueError
            If the path has an extension that is not ".csv".

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.to_csv_file("./src/resources/to_csv_file.csv")
        """
        path = _normalize_and_check_file_path(path, ".csv", [".csv"])
        path.parent.mkdir(parents=True, exist_ok=True)

        self._lazy_frame.sink_csv(path)

    def to_dict(self) -> dict[str, list[Any]]:
        """
        Return a dictionary that maps column names to column values.

        Returns
        -------
        dict_:
            Dictionary representation of the table.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.to_dict()
        {'a': [1, 2, 3], 'b': [4, 5, 6]}
        """
        return self._data_frame.to_dict(as_series=False)

    def to_json_file(
        self,
        path: str | Path,
        *,
        orientation: Literal["column", "row"] = "column",
    ) -> None:
        """
        Write the table to a JSON file.

        If the file and/or the parent directories do not exist, they will be created. If the file exists already, it
        will be overwritten.

        **Note:** This operation must fully load the data into memory, which can be expensive.

        Parameters
        ----------
        path:
            The path to the JSON file. If the file extension is omitted, it is assumed to be ".json".
        orientation:
            The orientation of the JSON file. If "column", the JSON file will be structured as a list of columns. If
            "row", the JSON file will be structured as a list of rows. Row orientation is more human-readable, but
            slower and less memory-efficient.

        Raises
        ------
        ValueError
            If the path has an extension that is not ".json".

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.to_json_file("./src/resources/to_json_file.json")
        """
        path = _normalize_and_check_file_path(path, ".json", [".json"])
        path.parent.mkdir(parents=True, exist_ok=True)

        # Write JSON to file
        self._data_frame.write_json(path, row_oriented=(orientation == "row"))

    def to_parquet_file(self, path: str | Path) -> None:
        """
        Write the table to a Parquet file.

        If the file and/or the parent directories do not exist, they will be created. If the file exists already, it
        will be overwritten.

        Parameters
        ----------
        path:
            The path to the Parquet file. If the file extension is omitted, it is assumed to be ".parquet".

        Raises
        ------
        ValueError
            If the path has an extension that is not ".parquet".

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"a": [1, 2, 3], "b": [4, 5, 6]})
        >>> table.to_parquet_file("./src/resources/to_parquet_file.parquet")
        """
        path = _normalize_and_check_file_path(path, ".parquet", [".parquet"])
        path.parent.mkdir(parents=True, exist_ok=True)

        self._lazy_frame.sink_parquet(path)

    def to_tabular_dataset(self, target_name: str, *, extra_names: list[str] | None = None) -> TabularDataset:
        """
        Return a new `TabularDataset` with columns marked as a target, feature, or extra.

        - The target column is the column that a model should predict.
        - Feature columns are columns that a model should use to make predictions.
        - Extra columns are columns that are neither feature nor target. They can be used to provide additional context,
          like an ID column.

        Feature columns are implicitly defined as all columns except the target and extra columns. If no extra columns
        are specified, all columns except the target column are used as features.

        Parameters
        ----------
        target_name:
            The name of the target column.
        extra_names:
            Names of the columns that are neither feature nor target. If None, no extra columns are used, i.e. all but
            the target column are used as features.

        Returns
        -------
        dataset:
            A new tabular dataset with the given target and feature names.

        Raises
        ------
        ValueError
            If the target column is also a feature column.
        ValueError
            If no feature columns are specified.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table(
        ...     {
        ...         "item": ["apple", "milk", "beer"],
        ...         "price": [1.10, 1.19, 1.79],
        ...         "amount_bought": [74, 72, 51],
        ...     }
        ... )
        >>> dataset = table.to_tabular_dataset(target_name="amount_bought", extra_names=["item"])
        """
        from safeds.data.labeled.containers import TabularDataset  # circular import

        return TabularDataset(
            self,
            target_name=target_name,
            extra_names=extra_names,
        )

    def to_time_series_dataset(
        self,
        target_name: str,
        window_size: int,
        *,
        extra_names: list[str] | None = None,
        forecast_horizon: int = 1,
        continuous: bool = False,
    ) -> TimeSeriesDataset:
        """
        Return a new `TimeSeriesDataset` with columns marked as a target column, time or feature columns.

        The original table is not modified.

        Parameters
        ----------
        target_name:
            The name of the target column.
        window_size:
            The number of consecutive sample to use as input for prediction.
        extra_names:
            Names of the columns that are neither features nor target. If None, no extra columns are used, i.e. all but
            the target column are used as features.
        forecast_horizon:
            The number of time steps to predict into the future.

        Returns
        -------
        dataset:
            A new time series dataset with the given target and feature names.

        Raises
        ------
        ValueError
            If the target column is also a feature column.
        ValueError
            If the time column is also a feature column.

        Examples
        --------
        >>> from safeds.data.tabular.containers import Table
        >>> table = Table({"day": [0, 1, 2], "price": [1.10, 1.19, 1.79], "amount_bought": [74, 72, 51]})
        >>> dataset = table.to_time_series_dataset(target_name="amount_bought", window_size=2)
        """
        from safeds.data.labeled.containers import TimeSeriesDataset  # circular import

        return TimeSeriesDataset(
            self,
            target_name=target_name,
            window_size=window_size,
            extra_names=extra_names,
            forecast_horizon=forecast_horizon,
            continuous=continuous,
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Dataframe interchange protocol
    # ------------------------------------------------------------------------------------------------------------------

    def __dataframe__(self, nan_as_null: bool = False, allow_copy: bool = True):  # type: ignore[no-untyped-def]
        """
        Return a dataframe object that conforms to the dataframe interchange protocol.

        Generally, there is no reason to call this method directly. The dataframe interchange protocol is designed to
        allow libraries to consume tabular data from different sources, such as `pandas` or `polars`. If you still
        decide to call this method, you should not rely on any capabilities of the returned object beyond the dataframe
        interchange protocol.

        The specification of the dataframe interchange protocol can be found
        [here](https://data-apis.org/dataframe-protocol/latest/index.html).

        **Note:** This operation must fully load the data into memory, which can be expensive.

        Parameters
        ----------
        nan_as_null:
            This parameter is deprecated and will be removed in a later revision of the dataframe interchange protocol.
            Setting it has no effect.
        allow_copy:
            Whether memory may be copied to create the dataframe object.

        Returns
        -------
        dataframe:
            A dataframe object that conforms to the dataframe interchange protocol.
        """
        return self._data_frame.__dataframe__(allow_copy=allow_copy)

    # ------------------------------------------------------------------------------------------------------------------
    # IPython integration
    # ------------------------------------------------------------------------------------------------------------------

    def _repr_html_(self) -> str:
        """
        Return a compact HTML representation of the table for IPython.

        **Note:** This operation must fully load the data into memory, which can be expensive.

        Returns
        -------
        html:
            The generated HTML.
        """
        return self._data_frame._repr_html_()

    # ------------------------------------------------------------------------------------------------------------------
    # Internal
    # ------------------------------------------------------------------------------------------------------------------

    # TODO
    def _into_dataloader(self, batch_size: int) -> DataLoader:
        """
        Return a Dataloader for the data stored in this table, used for predicting with neural networks.

        The original table is not modified.

        Parameters
        ----------
        batch_size:
            The size of data batches that should be loaded at one time.

        Returns
        -------
        result:
            The DataLoader.

        """
        import polars as pl
        import torch
        from torch.utils.data import DataLoader

        _init_default_device()

        return DataLoader(
            dataset=_create_dataset(self._data_frame.to_torch(dtype=pl.Float32).to(_get_device())),
            batch_size=batch_size,
            generator=torch.Generator(device=_get_device()),
        )


# TODO
def _create_dataset(features: Tensor) -> Dataset:
    from torch.utils.data import Dataset

    _init_default_device()

    class _CustomDataset(Dataset):
        def __init__(self, features: Tensor):
            self.X = features
            self.len = self.X.shape[0]

        def __getitem__(self, item: int) -> torch.Tensor:
            return self.X[item]

        def __len__(self) -> int:
            return self.len

    return _CustomDataset(features)



"""Layers of neural networks."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._convolutional2d_layer import Convolutional2DLayer, ConvolutionalTranspose2DLayer
    from ._dropout_layer import DropoutLayer
    from ._flatten_layer import FlattenLayer
    from ._forward_layer import ForwardLayer
    from ._gru_layer import GRULayer
    from ._layer import Layer
    from ._lstm_layer import LSTMLayer
    from ._pooling2d_layer import AveragePooling2DLayer, MaxPooling2DLayer

apipkg.initpkg(
    __name__,
    {
        "Convolutional2DLayer": "._convolutional2d_layer:Convolutional2DLayer",
        "ConvolutionalTranspose2DLayer": "._convolutional2d_layer:ConvolutionalTranspose2DLayer",
        "FlattenLayer": "._flatten_layer:FlattenLayer",
        "ForwardLayer": "._forward_layer:ForwardLayer",
        "Layer": "._layer:Layer",
        "LSTMLayer": "._lstm_layer:LSTMLayer",
        "GRULayer": "._gru_layer:GRULayer",
        "AveragePooling2DLayer": "._pooling2d_layer:AveragePooling2DLayer",
        "MaxPooling2DLayer": "._pooling2d_layer:MaxPooling2DLayer",
        "DropoutLayer": "._dropout_layer:DropoutLayer",
    },
)

__all__ = [
    "Convolutional2DLayer",
    "ConvolutionalTranspose2DLayer",
    "FlattenLayer",
    "ForwardLayer",
    "Layer",
    "LSTMLayer",
    "GRULayer",
    "AveragePooling2DLayer",
    "MaxPooling2DLayer",
    "DropoutLayer",
]



from __future__ import annotations

import copy
import multiprocessing as mp
from concurrent.futures import ALL_COMPLETED, ProcessPoolExecutor, wait
from typing import TYPE_CHECKING, Any, Generic, Literal, Self, TypeVar

from safeds._config import _init_default_device
from safeds._validation import _check_bounds, _ClosedBound
from safeds.data.image.containers import ImageList
from safeds.data.labeled.containers import ImageDataset, TabularDataset, TimeSeriesDataset
from safeds.data.labeled.containers._image_dataset import _ColumnAsTensor
from safeds.data.tabular.containers import Table
from safeds.data.tabular.transformation import OneHotEncoder
from safeds.exceptions import (
    FeatureDataMismatchError,
    FittingWithChoiceError,
    FittingWithoutChoiceError,
    InvalidModelStructureError,
    LearningError,
    ModelNotFittedError,
)
from safeds.ml.metrics import ClassificationMetrics, RegressionMetrics
from safeds.ml.nn.converters import (
    InputConversionImageToColumn,
    InputConversionImageToImage,
    InputConversionImageToTable,
)
from safeds.ml.nn.converters._input_converter_image import _InputConversionImage
from safeds.ml.nn.layers import (
    Convolutional2DLayer,
    FlattenLayer,
    ForwardLayer,
)
from safeds.ml.nn.layers._pooling2d_layer import _Pooling2DLayer
from safeds.ml.nn.typing import ConstantImageSize, ModelImageSize, VariableImageSize

if TYPE_CHECKING:
    from collections.abc import Callable

    from torch import nn
    from torch.nn import Module
    from transformers.image_processing_utils import BaseImageProcessor

    from safeds.ml.nn.converters import InputConversion
    from safeds.ml.nn.layers import Layer

IFT = TypeVar("IFT", TabularDataset, TimeSeriesDataset, ImageDataset)  # InputFitType
IPT = TypeVar("IPT", Table, TimeSeriesDataset, ImageList)  # InputPredictType


class NeuralNetworkRegressor(Generic[IFT, IPT]):
    """
    A NeuralNetworkRegressor is a neural network that is used for regression tasks.

    Parameters
    ----------
    input_conversion:
        to convert the input data for the neural network
    layers:
        a list of layers for the neural network to learn

    Raises
    ------
    InvalidModelStructureError
        if the defined model structure is invalid
    """

    def __init__(
        self,
        input_conversion: InputConversion[IFT, IPT],
        layers: list[Layer],
    ):
        if len(layers) == 0:
            raise InvalidModelStructureError("You need to provide at least one layer to a neural network.")
        if isinstance(input_conversion, _InputConversionImage):
            # TODO: why is this limitation needed? we might want to output the probability that an image shows a certain
            #  object, which would be a 1-dimensional output.
            if isinstance(input_conversion, InputConversionImageToColumn | InputConversionImageToTable):
                raise InvalidModelStructureError(
                    "A NeuralNetworkRegressor cannot be used with images as input and 1-dimensional data as output.",
                )
            data_dimensions = 2
            for layer in layers:
                if data_dimensions == 2 and (isinstance(layer, Convolutional2DLayer | _Pooling2DLayer)):
                    continue
                elif data_dimensions == 2 and isinstance(layer, FlattenLayer):
                    data_dimensions = 1
                elif data_dimensions == 1 and isinstance(layer, ForwardLayer):
                    continue
                else:
                    raise InvalidModelStructureError(
                        (
                            "The 2-dimensional data has to be flattened before using a 1-dimensional layer."
                            if data_dimensions == 2
                            else "You cannot use a 2-dimensional layer with 1-dimensional data."
                        ),
                    )
            if data_dimensions == 1 and isinstance(input_conversion, InputConversionImageToImage):
                raise InvalidModelStructureError(
                    "The output data would be 1-dimensional but the provided output conversion uses 2-dimensional data.",
                )
        else:
            for layer in layers:
                if isinstance(layer, Convolutional2DLayer | FlattenLayer | _Pooling2DLayer):
                    raise InvalidModelStructureError("You cannot use a 2-dimensional layer with 1-dimensional data.")

        self._input_conversion: InputConversion[IFT, IPT] = input_conversion
        self._model: Module | None = None
        self._layers: list[Layer] = layers
        self._input_size: int | ModelImageSize | None = None
        self._batch_size = 1
        self._is_fitted = False
        self._total_number_of_batches_done = 0
        self._total_number_of_epochs_done = 0

    @staticmethod
    def load_pretrained_model(huggingface_repo: str) -> NeuralNetworkRegressor:  # pragma: no cover
        """
        Load a pretrained model from a [Huggingface repository](https://huggingface.co/models/).

        Parameters
        ----------
        huggingface_repo:
            the name of the huggingface repository

        Returns
        -------
        pretrained_model:
            the pretrained model as a NeuralNetworkRegressor
        """
        from transformers import (
            AutoConfig,
            AutoImageProcessor,
            AutoModelForImageToImage,
            PretrainedConfig,
            Swin2SRForImageSuperResolution,
            Swin2SRImageProcessor,
        )

        _init_default_device()

        config: PretrainedConfig = AutoConfig.from_pretrained(huggingface_repo)

        if config.model_type != "swin2sr":
            raise ValueError("This model is not supported")

        model: Swin2SRForImageSuperResolution = AutoModelForImageToImage.from_pretrained(huggingface_repo)

        image_processor: Swin2SRImageProcessor = AutoImageProcessor.from_pretrained(huggingface_repo)

        if hasattr(config, "num_channels"):
            input_size = VariableImageSize(image_processor.pad_size, image_processor.pad_size, config.num_channels)
        else:  # Should never happen due to model check
            raise ValueError("This model is not supported")  # pragma: no cover

        in_conversion = InputConversionImageToImage(input_size)

        network = NeuralNetworkRegressor.__new__(NeuralNetworkRegressor)
        network._input_conversion = in_conversion
        network._model = model
        network._input_size = input_size
        network._batch_size = 1
        network._is_fitted = True
        network._total_number_of_epochs_done = 0
        network._total_number_of_batches_done = 0

        return network

    def fit(
        self,
        train_data: IFT,
        epoch_size: int = 25,
        batch_size: int = 1,
        learning_rate: float = 0.001,
        callback_on_batch_completion: Callable[[int, float], None] | None = None,
        callback_on_epoch_completion: Callable[[int, float], None] | None = None,
    ) -> Self:
        """
        Train the neural network with given training data.

        The original model is not modified.

        Parameters
        ----------
        train_data:
            The data the network should be trained on.
        epoch_size:
            The number of times the training cycle should be done.
        batch_size:
            The size of data batches that should be loaded at one time.
        learning_rate:
            The learning rate of the neural network.
        callback_on_batch_completion:
            Function used to view metrics while training. Gets called after a batch is completed with the index of the
            last batch and the overall loss average.
        callback_on_epoch_completion:
            Function used to view metrics while training. Gets called after an epoch is completed with the index of the
            last epoch and the overall loss average.

        Returns
        -------
        trained_model:
            The trained Model

        Raises
        ------
        OutOfBoundsError
            If epoch_size < 1
            If batch_size < 1
        """
        import torch
        from torch import nn

        from ._internal_model import _InternalModel  # Slow import on global level

        _init_default_device()

        if self._contains_choices():
            raise FittingWithChoiceError

        if not self._input_conversion._is_fit_data_valid(train_data):
            raise FeatureDataMismatchError

        _check_bounds("epoch_size", epoch_size, lower_bound=_ClosedBound(1))
        _check_bounds("batch_size", batch_size, lower_bound=_ClosedBound(1))

        copied_model = copy.deepcopy(self)
        # TODO: How is this supposed to work with pre-trained models? Should the old weights be kept or discarded?
        copied_model._model = _InternalModel(self._input_conversion, self._layers, is_for_classification=False)
        copied_model._input_size = copied_model._model.input_size
        copied_model._batch_size = batch_size

        # TODO: Re-enable or remove depending on how the above TODO is resolved
        # if copied_model._input_conversion._data_size != copied_model._input_size:
        #     raise InputSizeError(copied_model._input_conversion._data_size, copied_model._input_size)

        dataloader = copied_model._input_conversion._data_conversion_fit(train_data, copied_model._batch_size)

        loss_fn = nn.MSELoss()

        optimizer = torch.optim.SGD(copied_model._model.parameters(), lr=learning_rate)
        for _ in range(epoch_size):
            loss_sum = 0.0
            amount_of_loss_values_calculated = 0
            for x, y in iter(dataloader):
                optimizer.zero_grad()
                pred = copied_model._model(x)
                loss = loss_fn(pred, y)
                loss_sum += loss.item()
                amount_of_loss_values_calculated += 1
                loss.backward()
                optimizer.step()
                copied_model._total_number_of_batches_done += 1
                if callback_on_batch_completion is not None:
                    callback_on_batch_completion(
                        copied_model._total_number_of_batches_done,
                        loss_sum / amount_of_loss_values_calculated,
                    )
            copied_model._total_number_of_epochs_done += 1
            if callback_on_epoch_completion is not None:
                callback_on_epoch_completion(
                    copied_model._total_number_of_epochs_done,
                    loss_sum / amount_of_loss_values_calculated,
                )
        copied_model._is_fitted = True
        copied_model._model.eval()
        return copied_model

    def fit_by_exhaustive_search(
        self,
        train_data: IFT,
        optimization_metric: Literal[
            "mean_squared_error",
            "mean_absolute_error",
            "median_absolute_deviation",
            "coefficient_of_determination",
        ],
        epoch_size: int = 25,
        batch_size: int = 1,
        learning_rate: float = 0.001,
    ) -> Self:
        """
        Use the hyperparameter choices to create multiple models and fit them.

        **Note:** This model is not modified.

        Parameters
        ----------
        train_data:
            The data the network should be trained on.
        optimization_metric:
            The metric that should be used for determining the performance of a model.
        epoch_size:
            The number of times the training cycle should be done.
        batch_size:
            The size of data batches that should be loaded at one time.
        learning_rate:
            The learning rate of the neural network.

        Returns
        -------
        best_model:
            The model that performed the best out of all possible models given the Choices of hyperparameters.

        Raises
        ------
        FittingWithoutChoiceError
            When calling this method on a model without hyperparameter choices.
        LearningError
            If the training data contains invalid values or if the training failed. Currently raised, when calling this on RNNs or CNNs as well.
        """
        _init_default_device()

        if not self._contains_choices():
            raise FittingWithoutChoiceError

        if isinstance(train_data, TimeSeriesDataset):
            raise LearningError("RNN-Hyperparameter optimization is currently not supported.")  # pragma: no cover
        if isinstance(train_data, ImageDataset):
            raise LearningError("CNN-Hyperparameter optimization is currently not supported.")  # pragma: no cover

        _check_bounds("epoch_size", epoch_size, lower_bound=_ClosedBound(1))
        _check_bounds("batch_size", batch_size, lower_bound=_ClosedBound(1))

        list_of_models = self._get_models_for_all_choices()
        list_of_fitted_models: list[Self] = []

        with ProcessPoolExecutor(max_workers=len(list_of_models), mp_context=mp.get_context("spawn")) as executor:
            futures = []
            for model in list_of_models:
                futures.append(executor.submit(model.fit, train_data, epoch_size, batch_size, learning_rate))
            [done, _] = wait(futures, return_when=ALL_COMPLETED)
            for future in done:
                list_of_fitted_models.append(future.result())
        executor.shutdown()

        # Cross Validation
        [train_split, test_split] = train_data.to_table().split_rows(0.75)
        train_data = train_split.to_tabular_dataset(
            target_name=train_data.target.name,
            extra_names=train_data.extras.column_names,
        )
        test_data = test_split.to_tabular_dataset(
            target_name=train_data.target.name,
            extra_names=train_data.extras.column_names,
        ).features
        target_col = train_data.target

        best_model = None
        best_metric_value = None
        for fitted_model in list_of_fitted_models:
            if best_model is None:
                best_model = fitted_model
                match optimization_metric:
                    case "mean_squared_error":
                        best_metric_value = RegressionMetrics.mean_squared_error(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
                    case "mean_absolute_error":
                        best_metric_value = RegressionMetrics.mean_absolute_error(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
                    case "median_absolute_deviation":
                        best_metric_value = RegressionMetrics.median_absolute_deviation(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
                    case "coefficient_of_determination":
                        best_metric_value = RegressionMetrics.coefficient_of_determination(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
            else:
                match optimization_metric:
                    case "mean_squared_error":
                        error_of_fitted_model = RegressionMetrics.mean_squared_error(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
                        if error_of_fitted_model < best_metric_value:
                            best_model = fitted_model  # pragma: no cover
                            best_metric_value = error_of_fitted_model  # pragma: no cover
                    case "mean_absolute_error":
                        error_of_fitted_model = RegressionMetrics.mean_absolute_error(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
                        if error_of_fitted_model < best_metric_value:
                            best_model = fitted_model  # pragma: no cover
                            best_metric_value = error_of_fitted_model  # pragma: no cover
                    case "median_absolute_deviation":
                        error_of_fitted_model = RegressionMetrics.median_absolute_deviation(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
                        if error_of_fitted_model < best_metric_value:
                            best_model = fitted_model  # pragma: no cover
                            best_metric_value = error_of_fitted_model  # pragma: no cover
                    case "coefficient_of_determination":
                        error_of_fitted_model = RegressionMetrics.coefficient_of_determination(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
                        if error_of_fitted_model > best_metric_value:
                            best_model = fitted_model  # pragma: no cover
                            best_metric_value = error_of_fitted_model  # pragma: no cover
        assert best_model is not None  # just for linter
        best_model._is_fitted = True
        return best_model

    def _get_models_for_all_choices(self) -> list[Self]:

        all_possible_layer_combinations: list[list] = [[]]
        for layer in self._layers:
            if not layer._contains_choices():
                for item in all_possible_layer_combinations:
                    item.append(layer)
            else:
                updated_combinations = []
                versions_of_one_layer = layer._get_layers_for_all_choices()
                for version in versions_of_one_layer:
                    copy_of_all_current_possible_combinations = copy.deepcopy(all_possible_layer_combinations)
                    for combination in copy_of_all_current_possible_combinations:
                        combination.append(version)
                        updated_combinations.append(combination)
                all_possible_layer_combinations = updated_combinations

        models = []
        for combination in all_possible_layer_combinations:
            new_model = NeuralNetworkRegressor(input_conversion=self._input_conversion, layers=combination)
            models.append(new_model)
        return models  # type: ignore[return-value]

    def predict(self, test_data: IPT) -> IFT:
        """
        Make a prediction for the given test data.

        The original Model is not modified.

        Parameters
        ----------
        test_data:
            The data the network should predict.

        Returns
        -------
        prediction:
            The given test_data with an added "prediction" column at the end

        Raises
        ------
        ModelNotFittedError
            If the model has not been fitted yet
        """
        import torch

        _init_default_device()

        if not self._is_fitted or self._model is None:
            raise ModelNotFittedError
        if not self._input_conversion._is_predict_data_valid(test_data):
            raise FeatureDataMismatchError
        dataloader = self._input_conversion._data_conversion_predict(test_data, self._batch_size)
        predictions = []
        with torch.no_grad():
            for x in dataloader:
                elem = self._model(x)
                if not isinstance(elem, torch.Tensor) and hasattr(elem, "reconstruction"):
                    elem = elem.reconstruction  # pragma: no cover
                elif not isinstance(elem, torch.Tensor):
                    raise ValueError(f"Output of model has unsupported type: {type(elem)}")  # pragma: no cover
                predictions.append(elem.squeeze(dim=1))
        return self._input_conversion._data_conversion_output(
            test_data,
            torch.cat(predictions, dim=0),
        )

    @property
    def is_fitted(self) -> bool:
        """Whether the model is fitted."""
        return self._is_fitted

    @property
    def input_size(self) -> int | ModelImageSize | None:
        """The input size of the model."""
        # TODO: raise if not fitted, don't return None
        return self._input_size

    def _contains_choices(self) -> bool:
        """Whether the model contains choices in any layer."""
        return any(layer._contains_choices() for layer in self._layers)


class NeuralNetworkClassifier(Generic[IFT, IPT]):
    """
    A NeuralNetworkClassifier is a neural network that is used for classification tasks.

    Parameters
    ----------
    input_conversion:
        to convert the input data for the neural network
    layers:
        a list of layers for the neural network to learn

    Raises
    ------
    InvalidModelStructureError
        if the defined model structure is invalid
    """

    def __init__(
        self,
        input_conversion: InputConversion[IFT, IPT],
        layers: list[Layer],
    ):
        if len(layers) == 0:
            raise InvalidModelStructureError("You need to provide at least one layer to a neural network.")
        if isinstance(input_conversion, InputConversionImageToImage):
            raise InvalidModelStructureError("A NeuralNetworkClassifier cannot be used with images as output.")
        if isinstance(input_conversion, _InputConversionImage) and isinstance(
            input_conversion._input_size,
            VariableImageSize,
        ):
            raise InvalidModelStructureError(
                "A NeuralNetworkClassifier cannot be used with a InputConversionImage that uses a VariableImageSize.",
            )
        elif isinstance(input_conversion, _InputConversionImage):
            data_dimensions = 2
            for layer in layers:
                if data_dimensions == 2 and (isinstance(layer, Convolutional2DLayer | _Pooling2DLayer)):
                    continue
                elif data_dimensions == 2 and isinstance(layer, FlattenLayer):
                    data_dimensions = 1
                elif data_dimensions == 1 and isinstance(layer, ForwardLayer):
                    continue
                else:
                    raise InvalidModelStructureError(
                        (
                            "The 2-dimensional data has to be flattened before using a 1-dimensional layer."
                            if data_dimensions == 2
                            else "You cannot use a 2-dimensional layer with 1-dimensional data."
                        ),
                    )
            if data_dimensions == 2 and (
                isinstance(input_conversion, InputConversionImageToColumn | InputConversionImageToTable)
            ):
                raise InvalidModelStructureError(
                    "The output data would be 2-dimensional but the provided output conversion uses 1-dimensional data.",
                )
        else:
            for layer in layers:
                if isinstance(layer, Convolutional2DLayer | FlattenLayer | _Pooling2DLayer):
                    raise InvalidModelStructureError("You cannot use a 2-dimensional layer with 1-dimensional data.")

        self._input_conversion: InputConversion[IFT, IPT] = input_conversion
        self._model: nn.Module | None = None
        self._layers: list[Layer] = layers
        self._input_size: int | ModelImageSize | None = None
        self._batch_size = 1
        self._is_fitted = False
        self._num_of_classes = (
            layers[-1].output_size if isinstance(layers[-1].output_size, int) else -1
        )  # Is always int but linter doesn't know
        self._total_number_of_batches_done = 0
        self._total_number_of_epochs_done = 0

    @staticmethod
    def load_pretrained_model(huggingface_repo: str) -> NeuralNetworkClassifier:  # pragma: no cover
        """
        Load a pretrained model from a [Huggingface repository](https://huggingface.co/models/).

        Parameters
        ----------
        huggingface_repo:
            the name of the huggingface repository

        Returns
        -------
        pretrained_model:
            the pretrained model as a NeuralNetworkClassifier
        """
        from transformers import AutoConfig, AutoImageProcessor, AutoModelForImageClassification, PretrainedConfig
        from transformers.models.auto.modeling_auto import MODEL_FOR_IMAGE_CLASSIFICATION_MAPPING_NAMES

        _init_default_device()

        config: PretrainedConfig = AutoConfig.from_pretrained(huggingface_repo)

        if config.model_type not in MODEL_FOR_IMAGE_CLASSIFICATION_MAPPING_NAMES:
            raise ValueError("This model is not supported")

        model: Module = AutoModelForImageClassification.from_pretrained(huggingface_repo)

        image_processor: BaseImageProcessor = AutoImageProcessor.from_pretrained(huggingface_repo)
        if hasattr(image_processor, "size") and hasattr(config, "num_channels"):
            if "shortest_edge" in image_processor.size:
                input_size = ConstantImageSize(
                    image_processor.size.get("shortest_edge"),
                    image_processor.size.get("shortest_edge"),
                    config.num_channels,
                )
            else:
                input_size = ConstantImageSize(
                    image_processor.size.get("width"),
                    image_processor.size.get("height"),
                    config.num_channels,
                )
        else:  # Should never happen due to model check
            raise ValueError("This model is not supported")  # pragma: no cover

        label_dict: dict[str, str] = config.id2label
        column_name = "label"
        labels_table = Table({column_name: [label for _, label in label_dict.items()]})
        one_hot_encoder = OneHotEncoder(column_names=[column_name]).fit(labels_table)

        in_conversion = InputConversionImageToColumn(input_size)

        in_conversion._column_name = column_name
        in_conversion._one_hot_encoder = one_hot_encoder
        in_conversion._input_size = input_size
        in_conversion._output_type = _ColumnAsTensor
        num_of_classes = labels_table.row_count

        network = NeuralNetworkClassifier.__new__(NeuralNetworkClassifier)
        network._input_conversion = in_conversion
        network._model = model
        network._input_size = input_size
        network._batch_size = 1
        network._is_fitted = True
        network._num_of_classes = num_of_classes
        network._total_number_of_epochs_done = 0
        network._total_number_of_batches_done = 0

        return network

    def fit(
        self,
        train_data: IFT,
        epoch_size: int = 25,
        batch_size: int = 1,
        learning_rate: float = 0.001,
        callback_on_batch_completion: Callable[[int, float], None] | None = None,
        callback_on_epoch_completion: Callable[[int, float], None] | None = None,
    ) -> Self:
        """
        Train the neural network with given training data.

        The original model is not modified.

        Parameters
        ----------
        train_data:
            The data the network should be trained on.
        epoch_size:
            The number of times the training cycle should be done.
        batch_size:
            The size of data batches that should be loaded at one time.
        learning_rate:
            The learning rate of the neural network.
        callback_on_batch_completion:
            Function used to view metrics while training. Gets called after a batch is completed with the index of the
            last batch and the overall loss average.
        callback_on_epoch_completion:
            Function used to view metrics while training. Gets called after an epoch is completed with the index of the
            last epoch and the overall loss average.

        Returns
        -------
        trained_model:
            The trained Model

        Raises
        ------
        ValueError
            If epoch_size < 1
            If batch_size < 1
        """
        import torch
        from torch import nn

        from ._internal_model import _InternalModel  # Slow import on global level

        _init_default_device()

        if self._contains_choices():
            raise FittingWithChoiceError

        if not self._input_conversion._is_fit_data_valid(train_data):
            raise FeatureDataMismatchError

        _check_bounds("epoch_size", epoch_size, lower_bound=_ClosedBound(1))
        _check_bounds("batch_size", batch_size, lower_bound=_ClosedBound(1))

        copied_model = copy.deepcopy(self)
        # TODO: How is this supposed to work with pre-trained models? Should the old weights be kept or discarded?
        copied_model._model = _InternalModel(self._input_conversion, self._layers, is_for_classification=True)
        copied_model._batch_size = batch_size
        copied_model._input_size = copied_model._model.input_size

        # TODO: Re-enable or remove depending on how the above TODO is resolved
        # if copied_model._input_conversion._data_size != copied_model._input_size:
        #     raise InputSizeError(copied_model._input_conversion._data_size, copied_model._input_size)

        dataloader = copied_model._input_conversion._data_conversion_fit(
            train_data,
            copied_model._batch_size,
            copied_model._num_of_classes,
        )

        if copied_model._num_of_classes > 1:
            loss_fn = nn.CrossEntropyLoss()
        else:
            loss_fn = nn.BCELoss()

        optimizer = torch.optim.SGD(copied_model._model.parameters(), lr=learning_rate)
        for _ in range(epoch_size):
            loss_sum = 0.0
            amount_of_loss_values_calculated = 0
            for x, y in iter(dataloader):
                optimizer.zero_grad()
                pred = copied_model._model(x)

                loss = loss_fn(pred, y)
                loss_sum += loss.item()
                amount_of_loss_values_calculated += 1
                loss.backward()
                optimizer.step()

                copied_model._total_number_of_batches_done += 1
                if callback_on_batch_completion is not None:
                    callback_on_batch_completion(
                        copied_model._total_number_of_batches_done,
                        loss_sum / amount_of_loss_values_calculated,
                    )
            copied_model._total_number_of_epochs_done += 1
            if callback_on_epoch_completion is not None:
                callback_on_epoch_completion(
                    copied_model._total_number_of_epochs_done,
                    loss_sum / amount_of_loss_values_calculated,
                )
        copied_model._is_fitted = True
        copied_model._model.eval()
        return copied_model

    def fit_by_exhaustive_search(
        self,
        train_data: IFT,
        optimization_metric: Literal["accuracy", "precision", "recall", "f1_score"],
        positive_class: Any = None,
        epoch_size: int = 25,
        batch_size: int = 1,
        learning_rate: float = 0.001,
    ) -> Self:
        """
        Use the hyperparameter choices to create multiple models and fit them.

        **Note:** This model is not modified.

        Parameters
        ----------
        train_data:
            The data the network should be trained on.
        optimization_metric:
            The metric that should be used for determining the performance of a model.
        positive_class:
            The class to be considered positive. Only needs to be provided when choosing precision, recall or f1_score as the optimization metric.
        epoch_size:
            The number of times the training cycle should be done.
        batch_size:
            The size of data batches that should be loaded at one time.
        learning_rate:
            The learning rate of the neural network.

        Returns
        -------
        best_model:
            The model that performed the best out of all possible models given the Choices of hyperparameters.

        Raises
        ------
        FittingWithoutChoiceError
            When calling this method on a model without hyperparameter choices.
        LearningError
            If the training data contains invalid values or if the training failed. Currently raised, when calling this on RNNs or CNNs as well.
        """
        _init_default_device()

        if not self._contains_choices():
            raise FittingWithoutChoiceError

        if isinstance(train_data, TimeSeriesDataset):
            raise LearningError("RNN-Hyperparameter optimization is currently not supported.")  # pragma: no cover
        if isinstance(train_data, ImageDataset):
            raise LearningError("CNN-Hyperparameter optimization is currently not supported.")  # pragma: no cover

        _check_bounds("epoch_size", epoch_size, lower_bound=_ClosedBound(1))
        _check_bounds("batch_size", batch_size, lower_bound=_ClosedBound(1))

        list_of_models = self._get_models_for_all_choices()
        list_of_fitted_models: list[Self] = []

        with ProcessPoolExecutor(max_workers=len(list_of_models), mp_context=mp.get_context("spawn")) as executor:
            futures = []
            for model in list_of_models:
                futures.append(executor.submit(model.fit, train_data, epoch_size, batch_size, learning_rate))
            [done, _] = wait(futures, return_when=ALL_COMPLETED)
            for future in done:
                list_of_fitted_models.append(future.result())
        executor.shutdown()

        # Cross Validation
        [train_split, test_split] = train_data.to_table().split_rows(0.75)
        train_data = train_split.to_tabular_dataset(
            target_name=train_data.target.name,
            extra_names=train_data.extras.column_names,
        )
        test_data = test_split.to_tabular_dataset(
            target_name=train_data.target.name,
            extra_names=train_data.extras.column_names,
        ).features
        target_col = train_data.target

        best_model = None
        best_metric_value = None
        for fitted_model in list_of_fitted_models:
            if best_model is None:
                best_model = fitted_model
                match optimization_metric:
                    case "accuracy":
                        best_metric_value = ClassificationMetrics.accuracy(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
                    case "precision":
                        best_metric_value = ClassificationMetrics.precision(predicted=fitted_model.predict(test_data), expected=target_col, positive_class=positive_class)  # type: ignore[arg-type]
                    case "recall":
                        best_metric_value = ClassificationMetrics.recall(predicted=fitted_model.predict(test_data), expected=target_col, positive_class=positive_class)  # type: ignore[arg-type]
                    case "f1_score":
                        best_metric_value = ClassificationMetrics.f1_score(predicted=fitted_model.predict(test_data), expected=target_col, positive_class=positive_class)  # type: ignore[arg-type]
            else:
                match optimization_metric:
                    case "accuracy":
                        error_of_fitted_model = ClassificationMetrics.accuracy(predicted=fitted_model.predict(test_data), expected=target_col)  # type: ignore[arg-type]
                        if error_of_fitted_model > best_metric_value:
                            best_model = fitted_model  # pragma: no cover
                            best_metric_value = error_of_fitted_model  # pragma: no cover
                    case "precision":
                        error_of_fitted_model = ClassificationMetrics.precision(predicted=fitted_model.predict(test_data), expected=target_col, positive_class=positive_class)  # type: ignore[arg-type]
                        if error_of_fitted_model > best_metric_value:
                            best_model = fitted_model  # pragma: no cover
                            best_metric_value = error_of_fitted_model  # pragma: no cover
                    case "recall":
                        error_of_fitted_model = ClassificationMetrics.recall(predicted=fitted_model.predict(test_data), expected=target_col, positive_class=positive_class)  # type: ignore[arg-type]
                        if error_of_fitted_model > best_metric_value:
                            best_model = fitted_model  # pragma: no cover
                            best_metric_value = error_of_fitted_model  # pragma: no cover
                    case "f1_score":
                        error_of_fitted_model = ClassificationMetrics.f1_score(predicted=fitted_model.predict(test_data), expected=target_col, positive_class=positive_class)  # type: ignore[arg-type]
                        if error_of_fitted_model > best_metric_value:
                            best_model = fitted_model  # pragma: no cover
                            best_metric_value = error_of_fitted_model  # pragma: no cover
        assert best_model is not None  # just for linter
        best_model._is_fitted = True
        return best_model

    def _get_models_for_all_choices(self) -> list[Self]:

        all_possible_layer_combinations: list[list] = [[]]
        for layer in self._layers:
            if not layer._contains_choices():
                for item in all_possible_layer_combinations:
                    item.append(layer)
            else:
                updated_combinations = []
                versions_of_one_layer = layer._get_layers_for_all_choices()
                for version in versions_of_one_layer:
                    copy_of_all_current_possible_combinations = copy.deepcopy(all_possible_layer_combinations)
                    for combination in copy_of_all_current_possible_combinations:
                        combination.append(version)
                        updated_combinations.append(combination)
                all_possible_layer_combinations = updated_combinations

        models = []
        for combination in all_possible_layer_combinations:
            new_model = NeuralNetworkClassifier(input_conversion=self._input_conversion, layers=combination)
            models.append(new_model)
        return models  # type: ignore[return-value]

    def predict(self, test_data: IPT) -> IFT:
        """
        Make a prediction for the given test data.

        The original Model is not modified.

        Parameters
        ----------
        test_data:
            The data the network should predict.

        Returns
        -------
        prediction:
            The given test_data with an added "prediction" column at the end

        Raises
        ------
        ModelNotFittedError
            If the Model has not been fitted yet
        """
        import torch

        _init_default_device()

        if not self._is_fitted or self._model is None:
            raise ModelNotFittedError
        if not self._input_conversion._is_predict_data_valid(test_data):
            raise FeatureDataMismatchError
        dataloader = self._input_conversion._data_conversion_predict(test_data, self._batch_size)
        predictions = []
        with torch.no_grad():
            for x in dataloader:
                elem = self._model(x)
                if not isinstance(elem, torch.Tensor) and hasattr(elem, "logits"):
                    elem = elem.logits  # pragma: no cover
                elif not isinstance(elem, torch.Tensor):
                    raise ValueError(f"Output of model has unsupported type: {type(elem)}")  # pragma: no cover
                if self._num_of_classes > 1:
                    predictions.append(torch.argmax(elem, dim=1))
                else:
                    predictions.append(elem.squeeze(dim=1).round())
        return self._input_conversion._data_conversion_output(
            test_data,
            torch.cat(predictions, dim=0),
        )

    @property
    def is_fitted(self) -> bool:
        """Whether the model is fitted."""
        return self._is_fitted

    @property
    def input_size(self) -> int | ModelImageSize | None:
        """The input size of the model."""
        # TODO: raise if not fitted, don't return None
        return self._input_size

    def _contains_choices(self) -> bool:
        """Whether the model contains choices in any layer."""
        return any(layer._contains_choices() for layer in self._layers)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

In_co = TypeVar("In_co", covariant=True)
Out_co = TypeVar("Out_co", covariant=True)


class Dataset(Generic[In_co, Out_co], ABC):
    """A dataset is used as input to machine learning models."""

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def __eq__(self, other: object) -> bool: ...

    @abstractmethod
    def __hash__(self) -> int: ...

    @abstractmethod
    def __sizeof__(self) -> int: ...



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds.data.labeled.containers import TabularDataset
from safeds.data.tabular.containers import Column, Table
from safeds.exceptions import InvalidFitDataError

from ._input_converter import InputConversion

if TYPE_CHECKING:
    from torch import Tensor
    from torch.utils.data import DataLoader


class InputConversionTable(InputConversion[TabularDataset, Table]):
    """The input conversion for a neural network, defines the input parameters for the neural network."""

    def __init__(self) -> None:
        self._target_name = ""
        self._feature_names: list[str] = []
        self._first = True

    @property
    def _data_size(self) -> int:
        return len(self._feature_names)

    def _data_conversion_fit(self, input_data: TabularDataset, batch_size: int, num_of_classes: int = 1) -> DataLoader:
        return input_data._into_dataloader_with_classes(
            batch_size,
            num_of_classes,
        )

    def _data_conversion_predict(self, input_data: Table, batch_size: int) -> DataLoader:
        return input_data._into_dataloader(batch_size)

    def _data_conversion_output(self, input_data: Table, output_data: Tensor) -> TabularDataset:
        return input_data.add_columns([Column(self._target_name, output_data.tolist())]).to_tabular_dataset(
            self._target_name,
        )

    def _is_fit_data_valid(self, input_data: TabularDataset) -> bool:
        if self._first:
            self._feature_names = input_data.features.column_names
            self._target_name = input_data.target.name
            self._first = False

            columns_with_missing_values = []
            columns_with_non_numerical_data = []

            for col in input_data.features.add_columns([input_data.target]).to_columns():
                if col.missing_value_count() > 0:
                    columns_with_missing_values.append(col.name)
                if not col.type.is_numeric:
                    columns_with_non_numerical_data.append(col.name)

            reason = ""
            if len(columns_with_missing_values) > 0:
                reason += f"The following Columns contain missing values: {columns_with_missing_values}\n"
            if len(columns_with_non_numerical_data) > 0:
                reason += f"The following Columns contain non-numerical data: {columns_with_non_numerical_data}"
            if reason != "":
                raise InvalidFitDataError(reason)

        return (sorted(input_data.features.column_names)).__eq__(sorted(self._feature_names))

    def _is_predict_data_valid(self, input_data: Table) -> bool:
        return (sorted(input_data.column_names)).__eq__(sorted(self._feature_names))



"""Models for regression tasks."""

from typing import TYPE_CHECKING

import apipkg

if TYPE_CHECKING:
    from ._ada_boost_regressor import AdaBoostRegressor
    from ._arima import ArimaModelRegressor
    from ._baseline_regressor import BaselineRegressor
    from ._decision_tree_regressor import DecisionTreeRegressor
    from ._elastic_net_regressor import ElasticNetRegressor
    from ._gradient_boosting_regressor import GradientBoostingRegressor
    from ._k_nearest_neighbors_regressor import KNearestNeighborsRegressor
    from ._lasso_regressor import LassoRegressor
    from ._linear_regressor import LinearRegressor
    from ._random_forest_regressor import RandomForestRegressor
    from ._regressor import Regressor
    from ._ridge_regressor import RidgeRegressor
    from ._support_vector_regressor import SupportVectorRegressor

apipkg.initpkg(
    __name__,
    {
        "AdaBoostRegressor": "._ada_boost_regressor:AdaBoostRegressor",
        "ArimaModelRegressor": "._arima:ArimaModelRegressor",
        "BaselineRegressor": "._baseline_regressor:BaselineRegressor",
        "DecisionTreeRegressor": "._decision_tree_regressor:DecisionTreeRegressor",
        "ElasticNetRegressor": "._elastic_net_regressor:ElasticNetRegressor",
        "GradientBoostingRegressor": "._gradient_boosting_regressor:GradientBoostingRegressor",
        "KNearestNeighborsRegressor": "._k_nearest_neighbors_regressor:KNearestNeighborsRegressor",
        "LassoRegressor": "._lasso_regressor:LassoRegressor",
        "LinearRegressor": "._linear_regressor:LinearRegressor",
        "RandomForestRegressor": "._random_forest_regressor:RandomForestRegressor",
        "Regressor": "._regressor:Regressor",
        "RidgeRegressor": "._ridge_regressor:RidgeRegressor",
        "SupportVectorRegressor": "._support_vector_regressor:SupportVectorRegressor",
    },
)

__all__ = [
    "AdaBoostRegressor",
    "ArimaModelRegressor",
    "BaselineRegressor",
    "DecisionTreeRegressor",
    "ElasticNetRegressor",
    "GradientBoostingRegressor",
    "KNearestNeighborsRegressor",
    "LassoRegressor",
    "LinearRegressor",
    "RandomForestRegressor",
    "Regressor",
    "RidgeRegressor",
    "SupportVectorRegressor",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from safeds._utils import _structural_hash
from safeds._validation import _check_columns_exist
from safeds._validation._check_columns_are_numeric import _check_columns_are_numeric
from safeds.data.tabular.containers import Table
from safeds.exceptions import TransformerNotFittedError

from ._invertible_table_transformer import InvertibleTableTransformer

if TYPE_CHECKING:
    import polars as pl


class RangeScaler(InvertibleTableTransformer):
    """
    The RangeScaler transforms column values by scaling each value to a given range.

    Parameters
    ----------
    min_:
        The minimum of the new range after the transformation
    max_:
        The maximum of the new range after the transformation
    column_names:
        The list of columns used to fit the transformer. If `None`, all numeric columns are used.

    Raises
    ------
    ValueError
        If the given minimum is greater or equal to the given maximum
    """

    # ------------------------------------------------------------------------------------------------------------------
    # Dunder methods
    # ------------------------------------------------------------------------------------------------------------------

    def __init__(self, *, column_names: str | list[str] | None = None, min_: float = 0.0, max_: float = 1.0) -> None:
        super().__init__(column_names)

        if min_ >= max_:
            raise ValueError('Parameter "max_" must be greater than parameter "min_".')

        # Parameters
        self._min: float = min_
        self._max: float = max_

        # Internal state
        self._data_min: pl.DataFrame | None = None
        self._data_max: pl.DataFrame | None = None

    def __hash__(self) -> int:
        return _structural_hash(
            super().__hash__(),
            self._min,
            self._max,
            # Leave out the internal state for faster hashing
        )

    # ------------------------------------------------------------------------------------------------------------------
    # Properties
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def is_fitted(self) -> bool:
        """Whether the transformer is fitted."""
        return self._data_min is not None and self._data_max is not None

    @property
    def min(self) -> float:
        """The minimum of the new range after the transformation."""
        return self._min

    @property
    def max(self) -> float:
        """The maximum of the new range after the transformation."""
        return self._max

    # ------------------------------------------------------------------------------------------------------------------
    # Learning and transformation
    # ------------------------------------------------------------------------------------------------------------------

    def fit(self, table: Table) -> RangeScaler:
        """
        Learn a transformation for a set of columns in a table.

        This transformer is not modified.

        Parameters
        ----------
        table:
            The table used to fit the transformer.

        Returns
        -------
        fitted_transformer:
            The fitted transformer.

        Raises
        ------
        ColumnNotFoundError
            If column_names contain a column name that is missing in the table.
        ColumnTypeError
            If at least one of the specified columns in the table is not numeric.
        ValueError
            If the table contains 0 rows.
        """
        if self._column_names is None:
            column_names = [name for name in table.column_names if table.get_column_type(name).is_numeric]
        else:
            column_names = self._column_names
            _check_columns_exist(table, column_names)
            _check_columns_are_numeric(table, column_names, operation="fit a RangeScaler")

        if table.row_count == 0:
            raise ValueError("The RangeScaler cannot be fitted because the table contains 0 rows")

        # Learn the transformation
        _data_min = table._lazy_frame.select(column_names).min().collect()
        _data_max = table._lazy_frame.select(column_names).max().collect()

        # Create a copy with the learned transformation
        result = RangeScaler(min_=self._min, max_=self._max, column_names=column_names)
        result._data_min = _data_min
        result._data_max = _data_max

        return result

    def transform(self, table: Table) -> Table:
        """
        Apply the learned transformation to a table.

        The table is not modified.

        Parameters
        ----------
        table:
            The table to which the learned transformation is applied.

        Returns
        -------
        transformed_table:
            The transformed table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        ColumnNotFoundError
            If the input table does not contain all columns used to fit the transformer.
        ColumnTypeError
            If at least one of the columns in the input table that is used to fit contains non-numerical data.
        """
        import polars as pl

        # Used in favor of is_fitted, so the type checker is happy
        if self._column_names is None or self._data_min is None or self._data_max is None:
            raise TransformerNotFittedError

        _check_columns_exist(table, self._column_names)
        _check_columns_are_numeric(table, self._column_names, operation="transform with a RangeScaler")

        columns = [
            (
                (pl.col(name) - self._data_min.get_column(name))
                / (self._data_max.get_column(name) - self._data_min.get_column(name))
                * (self._max - self._min)
                + self._min
            )
            for name in self._column_names
        ]

        return Table._from_polars_lazy_frame(
            table._lazy_frame.with_columns(columns),
        )

    def inverse_transform(self, transformed_table: Table) -> Table:
        """
        Undo the learned transformation.

        The table is not modified.

        Parameters
        ----------
        transformed_table:
            The table to be transformed back to the original version.

        Returns
        -------
        original_table:
            The original table.

        Raises
        ------
        TransformerNotFittedError
            If the transformer has not been fitted yet.
        ColumnNotFoundError
            If the input table does not contain all columns used to fit the transformer.
        ColumnTypeError
            If the transformed columns of the input table contain non-numerical data.
        """
        import polars as pl

        # Used in favor of is_fitted, so the type checker is happy
        if self._column_names is None or self._data_min is None or self._data_max is None:
            raise TransformerNotFittedError

        _check_columns_exist(transformed_table, self._column_names)
        _check_columns_are_numeric(
            transformed_table,
            self._column_names,
            operation="inverse-transform with a RangeScaler",
        )

        columns = [
            (
                (pl.col(name) - self._min)
                / (self._max - self._min)
                * (self._data_max.get_column(name) - self._data_min.get_column(name))
                + self._data_min.get_column(name)
            )
            for name in self._column_names
        ]

        return Table._from_polars_lazy_frame(
            transformed_table._lazy_frame.with_columns(columns),
        )
